import logging
import sys
from imp import load_source
from os import listdir
from os.path import isfile, join, abspath

import bartender
from bartender.local_plugins.plugin_runner import LocalPluginRunner
from bg_utils.mongo.models import Instance, System


class LocalPluginLoader(object):
    """Class that helps with loading local plugins"""

    logger = logging.getLogger(__name__)

    def __init__(self, validator, registry):
        self.validator = validator
        self.registry = registry

    def load_plugins(self):
        """Load all plugins

        After each has been loaded, it checks the requirements to ensure
        the plugin can be loaded correctly.
        """
        for plugin_path in self.scan_plugin_path():
            try:
                self.load_plugin(plugin_path)
            except Exception as ex:
                self.logger.exception(
                    "Exception while loading plugin %s: %s", plugin_path, ex
                )

        self.validate_plugin_requirements()

    def scan_plugin_path(self, path=None):
        """Find valid plugin directories in a given path.

        Note: This scan does not walk the directory tree - all plugins must be
        in the top level of the given path.

        :param path: The path to scan for plugins. If none will default to the
            plugin path specified at initialization.

        :return: A list containing paths specifying plugins
        """
        path = path or bartender.config.plugin.local.directory

        if path is None:
            return []
        else:
            return [
                abspath(join(path, plugin))
                for plugin in listdir(path)
                if not isfile(join(path, plugin))
            ]

    def validate_plugin_requirements(self):
        """Validate requirements for each plugin can be satisfied"""
        plugin_list = self.registry.get_all_plugins()
        plugin_names = self.registry.get_unique_plugin_names()
        plugins_to_remove = []

        for plugin in plugin_list:
            for required_plugin in plugin.requirements:
                if required_plugin not in plugin_names:
                    self.logger.warning(
                        "Not loading plugin %s - plugin requirement %s is not "
                        "a known plugin.",
                        plugin.system.name,
                        required_plugin,
                    )
                    plugins_to_remove.append(plugin)

        for plugin in plugins_to_remove:
            self.registry.remove(plugin.unique_name)

    def load_plugin(self, plugin_path):
        """Loads a plugin given a path to a plugin directory.

        It will use the validator to validate the plugin before registering the
        plugin in the database as well as adding an entry to the plugin map.

        :param plugin_path: The path of the plugin to load
        :return: The loaded plugin
        """
        if not self.validator.validate_plugin(plugin_path):
            self.logger.warning(
                "Not loading plugin at %s because it was invalid.", plugin_path
            )
            return False

        config = self._load_plugin_config(join(plugin_path, "beer.conf"))

        plugin_name = config["NAME"]
        plugin_version = config["VERSION"]
        plugin_entry = config["PLUGIN_ENTRY"]
        plugin_instances = config["INSTANCES"]
        plugin_args = config["PLUGIN_ARGS"]

        # If this system already exists we need to do some stuff
        plugin_id = None
        plugin_commands = []
        # TODO: replace this with a call from the EasyClient
        # plugin_system = self.easy_client.find_unique_system(
        #     name=plugin_name, version=plugin_version)
        plugin_system = System.find_unique(plugin_name, plugin_version)

        if plugin_system:
            # Remove the current instances so they aren't left dangling
            # TODO: This should be replaced with a network call
            plugin_system.delete_instances()

            # Carry these over to the new system
            plugin_id = plugin_system.id
            plugin_commands = plugin_system.commands

        plugin_system = System(
            id=plugin_id,
            name=plugin_name,
            version=plugin_version,
            commands=plugin_commands,
            instances=[
                Instance(name=instance_name) for instance_name in plugin_instances
            ],
            max_instances=len(plugin_instances),
            description=config.get("DESCRIPTION"),
            icon_name=config.get("ICON_NAME"),
            display_name=config.get("DISPLAY_NAME"),
            metadata=config.get("METADATA"),
        )

        # TODO: Right now, we have to save this system because the LocalPluginRunner
        # uses the database to determine status, specifically, it calls reload on the
        # instance object which we need to change to satisfy
        plugin_system.deep_save()

        plugin_list = []
        plugin_log_directory = bartender.config.plugin.local.log_directory
        for instance_name in plugin_instances:
            plugin = LocalPluginRunner(
                plugin_entry,
                plugin_system,
                instance_name,
                abspath(plugin_path),
                bartender.config.web.host,
                bartender.config.web.port,
                bartender.config.web.ssl_enabled,
                plugin_args=plugin_args.get(instance_name),
                environment=config["ENVIRONMENT"],
                requirements=config["REQUIRES"],
                plugin_log_directory=plugin_log_directory,
                url_prefix=bartender.config.web.url_prefix,
                ca_verify=bartender.config.web.ca_verify,
                ca_cert=bartender.config.web.ca_cert,
                username=bartender.config.plugin.local.auth.username,
                password=bartender.config.plugin.local.auth.password,
                log_level=config["LOG_LEVEL"],
            )

            self.registry.register_plugin(plugin)
            plugin_list.append(plugin)

        return plugin_list

    def _load_plugin_config(self, path_to_config):
        """Loads a validated plugin config"""
        self.logger.debug("Loading configuration at %s", path_to_config)

        config_module = load_source("BGPLUGINCONFIG", path_to_config)

        instances = getattr(config_module, "INSTANCES", None)
        plugin_args = getattr(config_module, "PLUGIN_ARGS", None)
        user_log_level = getattr(config_module, "LOG_LEVEL", "INFO")
        log_level = self._convert_log_level(user_log_level)

        if instances is None and plugin_args is None:
            instances = ["default"]
            plugin_args = {"default": None}

        elif plugin_args is None:
            plugin_args = {}
            for instance_name in instances:
                plugin_args[instance_name] = None

        elif instances is None:
            if isinstance(plugin_args, list):
                instances = ["default"]
                plugin_args = {"default": plugin_args}
            elif isinstance(plugin_args, dict):
                instances = list(plugin_args.keys())
            else:
                raise ValueError("Unknown plugin args type: %s" % plugin_args)

        elif isinstance(plugin_args, list):
            temp_args = {}
            for instance_name in instances:
                temp_args[instance_name] = plugin_args

            plugin_args = temp_args

        config = {
            "NAME": config_module.NAME,
            "VERSION": config_module.VERSION,
            "INSTANCES": instances,
            "PLUGIN_ENTRY": config_module.PLUGIN_ENTRY,
            "PLUGIN_ARGS": plugin_args,
            "DESCRIPTION": getattr(config_module, "DESCRIPTION", ""),
            "ICON_NAME": getattr(config_module, "ICON_NAME", None),
            "DISPLAY_NAME": getattr(config_module, "DISPLAY_NAME", None),
            "REQUIRES": getattr(config_module, "REQUIRES", []),
            "ENVIRONMENT": getattr(config_module, "ENVIRONMENT", {}),
            "METADATA": getattr(config_module, "METADATA", {}),
            "LOG_LEVEL": log_level,
        }

        if "BGPLUGINCONFIG" in sys.modules:
            del sys.modules["BGPLUGINCONFIG"]

        return config

    def _convert_log_level(self, level_name):
        try:
            return getattr(logging, str(level_name).upper())
        except AttributeError:
            return logging.INFO



import logging
import logging.handlers
import os
import sys


class PluginHandler(object):
    """Basic Logging Handler for Plugins"""

    def __init__(self, handlerFactory, plugin_name, log_directory=None, **kw):
        if log_directory is not None:
            kw["filename"] = os.path.join(log_directory, plugin_name + ".log")

        kw["maxBytes"] = 10485760  # 10MB
        kw["backupCount"] = 5
        self._handler = handlerFactory(**kw)

    def __getattr__(self, attr):
        if hasattr(self._handler, attr):
            return getattr(self._handler, attr)
        raise AttributeError(attr)


# Since we are imitating the logging module, we will allow camel case method names
def getPluginLogger(
    name, format_string=None, log_directory=None, log_name=None, log_level=None
):
    """Get a logger for a plugin

    Args:
        name (str): The name of the logger to create
        format_string (str): The format_string to use with the logger
            If ``None`` then messages will be logged as-is
        log_directory (str, optional): Directory that will hold the log file
            If not given a logger for STDOUT will be constructed
        log_name (str, optional): The name of the log file
            * If ``log_directory`` is not provided this is not used
            * If not provided the ``name`` value will be used
        log_level (int, optional): Log level for this logger/handler pair.
            If none are set, then the current effective level will be used.

     Returns:
        The configured logger instance
    """
    if log_level is None:
        log_level = logging.getLogger(__name__).getEffectiveLevel()
    log = logging.getLogger(name)
    log.propagate = False
    if len(log.handlers) > 0:
        return log

    if log_directory:
        handler = PluginHandler(
            logging.handlers.RotatingFileHandler, log_name or name, log_directory
        )
    else:
        handler = logging.StreamHandler(sys.stdout)

    handler.setLevel(log_level)
    handler.setFormatter(logging.Formatter(format_string))
    log.addHandler(handler)
    log.setLevel(log_level)

    return log


def getLogLevels():
    try:
        # Python 2
        return [n for n in getattr(logging, "_levelNames").keys() if isinstance(n, str)]
    except AttributeError:
        # Python 3
        return [n for n in getattr(logging, "_nameToLevel").keys()]



class ConfigurationError(Exception):
    """Generic configuration error"""

    pass


class PluginValidationError(Exception):
    """Plugin could not be created successfully"""

    pass


class PluginStartupError(Exception):
    """Plugin could not be started"""

    pass


class ShutdownError(Exception):
    """Backend has been shut down"""

    pass



import logging
import logging.config

import bg_utils
import brewtils.rest
import bartender._version
from bartender.app import BartenderApp
from bartender.errors import ConfigurationError
from bartender.specification import get_default_logging_config
from brewtils.errors import ValidationError
from brewtils.rest.easy_client import EasyClient

__version__ = bartender._version.__version__

# COMPONENTS #
application = None
config = None
logger = None
bv_client = None


def setup_bartender(spec, cli_args):
    global application, config, logger, bv_client

    config = bg_utils.load_application_config(spec, cli_args)
    config.web.url_prefix = brewtils.rest.normalize_url_prefix(config.web.url_prefix)

    log_default = get_default_logging_config(config.log.level, config.log.file)
    bg_utils.setup_application_logging(config, log_default)
    logger = logging.getLogger(__name__)

    bv_client = EasyClient(**config.web)

    application = BartenderApp()

    logger.debug("Successfully loaded the bartender application")


def progressive_backoff(func, stoppable_thread, failure_message):
    wait_time = 0.1
    while not stoppable_thread.stopped() and not func():
        logger.warning(failure_message)
        logger.warning("Waiting %.1f seconds before next attempt", wait_time)

        stoppable_thread.wait(wait_time)
        wait_time = min(wait_time * 2, 30)


def ensure_admin():
    # Either brew-view auth must be disabled (anonymous user will have bg-all)
    # or the user must have bg-all permissions
    try:
        bartender_user = bv_client.who_am_i()
    except ValidationError:
        raise ConfigurationError(
            "Unable to authenticate using provided username and password. "
            "This usually indicates an incorrect password - please check the "
            "web.username and web.password fields in the configuration."
        )

    if "bg-all" not in bartender_user.permissions:
        if config.web.username:
            raise ConfigurationError(
                'User "%s" does not have "bg-all" permission. Please check '
                "your configuration (specifically web.username and "
                "web.password fields)" % config.web.username
            )
        else:
            raise ConfigurationError(
                "It appears that Brew-view is operating with authentication "
                "enabled and no username / password was provided. Please check "
                "your configuration (specifically web.username and "
                "web.password fields)."
            )



import logging
import sys
from imp import load_source
from os.path import isdir, isfile, join

from bartender.errors import PluginValidationError


class LocalPluginValidator(object):
    """Class Used for Validating Local Plugins"""

    CONFIG_NAME = "beer.conf"
    NAME_KEY = "NAME"
    VERSION_KEY = "VERSION"
    ENTRY_POINT_KEY = "PLUGIN_ENTRY"
    INSTANCES_KEY = "INSTANCES"
    ARGS_KEY = "PLUGIN_ARGS"
    REQUIRED_KEYS = [NAME_KEY, VERSION_KEY, ENTRY_POINT_KEY]

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def validate_plugin(self, path_to_plugin):
        """Validates a Plugin and its arguments

        :param path_to_plugin:
        :return:
        """
        try:
            self.logger.debug("Validating Plugin at %s ", path_to_plugin)

            self.validate_plugin_path(path_to_plugin)
            self.logger.debug("Validated Plugin Path successfully.")
            self.validate_plugin_config(path_to_plugin)
            self.logger.debug("Validated Plugin Config successfully.")

            self.logger.debug("Successfully validated Plugin at %s", path_to_plugin)
            return True

        except PluginValidationError as pve:
            self.logger.error(
                "Validation error occurred on plugin located at : %s", path_to_plugin
            )
            self.logger.error(str(pve))
            return False
        finally:
            if "BGPLUGINCONFIG" in sys.modules:
                del sys.modules["BGPLUGINCONFIG"]

    def validate_plugin_path(self, path_to_plugin):
        """Validates that a plugin path is actually a path and not a single file."""
        if path_to_plugin is None or not isdir(path_to_plugin):
            raise PluginValidationError(
                'Plugin path "%s" is not a directory' % path_to_plugin
            )

        return True

    def validate_plugin_config(self, path_to_plugin):
        """Validates that there is a beer.conf file in the path_to_plugin

        :param path_to_plugin: Path to the plugin
        :return: True if beer.conf exists
        :raises: PluginValidationError if beer.conf is not found
        """
        if path_to_plugin is None:
            raise PluginValidationError(
                "Attempted to validate plugin config, " "but the plugin_path is None."
            )

        path_to_config = join(path_to_plugin, self.CONFIG_NAME)

        if not isfile(path_to_config):
            raise PluginValidationError(
                "Could not validate config file. It does not exist."
            )

        config_module = self._load_plugin_config(path_to_plugin)
        self.validate_required_config_keys(config_module)
        self.logger.debug("Required keys are present.")
        self.validate_entry_point(config_module, path_to_plugin)
        self.logger.debug("Validated Plugin Entry Point successfully.")
        self.validate_instances_and_args(config_module)
        self.logger.debug("Validated plugin instances & arguments successfully.")
        self.validate_plugin_environment(config_module)
        self.logger.debug("Validated Plugin Environment successfully.")
        return True

    def validate_required_config_keys(self, config_module):
        if config_module is None:
            raise PluginValidationError("Configuration is None. This is not allowed.")

        for key in self.REQUIRED_KEYS:
            if not hasattr(config_module, key):
                raise PluginValidationError(
                    "Required key '%s' is not present. " "This is not allowed." % key
                )

    def _load_plugin_config(self, path_to_plugin):
        """Loads an already validated plugin Config.

        That is, it assumes path_to_config will exist and that it can load it succesfully"""
        path_to_config = join(path_to_plugin, self.CONFIG_NAME)

        return load_source("BGPLUGINCONFIG", path_to_config)

    def validate_entry_point(self, config_module, path_to_plugin):
        """Validates a plugin's entry point. Returns True or throws a PluginValidationError

        An entry point is considered valid if the config has an entry with key PLUGIN_ENTRY
        and the value is a path to either a file or the name of a runnable Python module.

        :param config_module: The previously loaded configuration for the plugin
        :param path_to_plugin: The path to the root of the plugin
        """
        if config_module is None:
            raise PluginValidationError("Configuration is None. This is not allowed.")

        if path_to_plugin is None:
            raise PluginValidationError("Path to Plugin is None. This is not allowed.")

        if not hasattr(config_module, self.ENTRY_POINT_KEY):
            raise PluginValidationError(
                "No %s defined in the plugin configuration." % self.ENTRY_POINT_KEY
            )

        entry_point = getattr(config_module, self.ENTRY_POINT_KEY)

        if isfile(join(path_to_plugin, entry_point)):
            return
        elif entry_point.startswith("-m "):
            pkg_path = join(path_to_plugin, entry_point[3:])
            if (
                isdir(pkg_path)
                and isfile(join(pkg_path, "__init__.py"))
                and isfile(join(pkg_path, "__main__.py"))
            ):
                return
        else:
            raise PluginValidationError(
                "The %s must be a Python script or a runnable Python package: %s"
                % (self.ENTRY_POINT_KEY, entry_point)
            )

    def validate_instances_and_args(self, config_module):
        if config_module is None:
            raise PluginValidationError("Configuration is None. This is not allowed.")

        plugin_args = getattr(config_module, self.ARGS_KEY, None)
        instances = getattr(config_module, self.INSTANCES_KEY, None)

        if instances is not None and not isinstance(instances, list):
            raise PluginValidationError(
                "'%s' entry was not None or a list. This is invalid. "
                "Got: %s" % (self.INSTANCES_KEY, instances)
            )

        if plugin_args is None:
            return True
        elif isinstance(plugin_args, list):
            return self.validate_individual_plugin_arguments(plugin_args)
        elif isinstance(plugin_args, dict):
            for instance_name, instance_args in plugin_args.items():
                if instances is not None and instance_name not in instances:
                    raise PluginValidationError(
                        "'%s' contains key '%s' but that instance is not specified in the '%s'"
                        "entry." % (self.ARGS_KEY, instance_name, self.INSTANCES_KEY)
                    )
                self.validate_individual_plugin_arguments(instance_args)

            if instances:
                for instance_name in instances:
                    if instance_name not in plugin_args.keys():
                        raise PluginValidationError(
                            "'%s' contains key '%s' but that instance is not specified in the "
                            "'%s' entry."
                            % (self.INSTANCES_KEY, instance_name, self.ARGS_KEY)
                        )

            return True
        else:
            raise PluginValidationError(
                "'%s' entry was not a list or dictionary. This is invalid. "
                "Got: %s" % (self.ARGS_KEY, plugin_args)
            )

    def validate_individual_plugin_arguments(self, plugin_args):
        """Validates an individual PLUGIN_ARGS entry"""

        if plugin_args is not None and not isinstance(plugin_args, list):
            self.logger.error(
                "Invalid Plugin Argument Specified. It was not a list or None. "
                "This is not allowed"
            )
            raise PluginValidationError(
                "Invalid Plugin Argument Specified: %s. It was not a "
                "list or None. This is not allowed." % plugin_args
            )

        if isinstance(plugin_args, list):
            for plugin_arg in plugin_args:
                if not isinstance(plugin_arg, str):
                    self.logger.error(
                        "Invalid plugin argument: %s - this argument must be a string",
                        plugin_arg,
                    )
                    raise PluginValidationError(
                        "Invalid plugin argument: %s - this argument must be a string."
                        % plugin_arg
                    )

        return True

    def validate_plugin_environment(self, config_module):
        """Validates ENVIRONMENT if specified.

        ENVIRONMENT must be a dictionary of Strings to Strings. Otherwise it is invalid.

        :param config_module:
        :return: True if valid
        :raises: PluginValidationError if something goes wrong while validating
        """
        if config_module is None:
            self.logger.error("Configuration is None. This is not allowed.")
            raise PluginValidationError("Configuration is None. This is not allowed.")

        if hasattr(config_module, "ENVIRONMENT"):
            env = config_module.ENVIRONMENT
            if not isinstance(env, dict):
                self.logger.error(
                    "Invalid ENVIRONMENT specified: %s. This argument must "
                    "be a dictionary.",
                    env,
                )
                raise PluginValidationError(
                    "Invalid ENVIRONMENT specified: %s. This argument "
                    "must be a dictionary." % env
                )

            for key, value in env.items():
                if not isinstance(key, str):
                    self.logger.error(
                        "Invalid Key: %s specified for plugin environment. "
                        "This must be a String.",
                        key,
                    )
                    raise PluginValidationError(
                        "Invalid Key: %s specified for plugin "
                        "environment. This must be a String." % key
                    )

                if key.startswith("BG_"):
                    self.logger.error(
                        "Invalid key: %s specified for plugin environment. The 'BG_' prefix is a "
                        "special case for beer-garden only environment variables. You will have to "
                        "pick another name. Sorry for the inconvenience." % key
                    )
                    raise PluginValidationError(
                        "Invalid key: %s specified for plugin environment. The 'BG_' prefix "
                        "is a special case for beer-garden only environment variables. You will "
                        "have to pick another name. Sorry for the inconvenience." % key
                    )

                if not isinstance(value, str):
                    self.logger.error(
                        "Invalid Key: %s specified for plugin environment. This must be a String.",
                        value,
                    )
                    raise PluginValidationError(
                        "Invalid Value: %s specified for plugin environment. This must be a "
                        "String." % value
                    )

        return True



import logging

from datetime import datetime, timedelta
from bg_utils.mongo.models import Instance, Request
from brewtils.stoppable_thread import StoppableThread


class PluginStatusMonitor(StoppableThread):
    """Monitor plugin heartbeats and update plugin status"""

    def __init__(self, clients, heartbeat_interval=10, timeout_seconds=30):
        self.logger = logging.getLogger(__name__)
        self.display_name = "Plugin Status Monitor"
        self.clients = clients
        self.heartbeat_interval = heartbeat_interval
        self.timeout = timedelta(seconds=timeout_seconds)
        self.status_request = Request(command="_status", command_type="EPHEMERAL")

        super(PluginStatusMonitor, self).__init__(
            logger=self.logger, name="PluginStatusMonitor"
        )

    def run(self):
        self.logger.info(self.display_name + " is started")

        while not self.wait(self.heartbeat_interval):
            self.request_status()
            self.check_status()

        self.logger.info(self.display_name + " is stopped")

    def request_status(self):
        try:
            self.clients["pika"].publish_request(
                self.status_request,
                routing_key="admin",
                expiration=str(self.heartbeat_interval * 1000),
            )
        except Exception as ex:
            self.logger.warning("Unable to publish status request: %s", str(ex))

    def check_status(self):
        """Update instance status if necessary"""

        for instance in Instance.objects:
            if self.stopped():
                break

            last_heartbeat = instance.status_info["heartbeat"]

            if last_heartbeat:
                if (
                    instance.status == "RUNNING"
                    and datetime.utcnow() - last_heartbeat >= self.timeout
                ):
                    instance.status = "UNRESPONSIVE"
                    instance.save()
                elif (
                    instance.status
                    in ["UNRESPONSIVE", "STARTING", "INITIALIZING", "UNKNOWN"]
                    and datetime.utcnow() - last_heartbeat < self.timeout
                ):
                    instance.status = "RUNNING"
                    instance.save()



import logging
from concurrent.futures import wait, ThreadPoolExecutor, ALL_COMPLETED
from threading import Event

from thriftpy2.server import TThreadedServer
from thriftpy2.thrift import TProcessor
from thriftpy2.transport import TServerSocket, TSSLServerSocket

import bg_utils
from brewtils.stoppable_thread import StoppableThread
import bartender


class BartenderThriftServer(TThreadedServer, StoppableThread):
    """Thrift server that uses a ThreadPoolExecutor to process requests"""

    # Amount of time (in seconds) after shutdown requested to wait for workers to finish processing
    WORKER_TIMEOUT = 5

    def __init__(self, *args, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.display_name = "Thrift Server"
        self.pool = ThreadPoolExecutor(max_workers=bartender.config.thrift.max_workers)
        self.futures = set()
        self.finished = Event()

        StoppableThread.__init__(
            self, logger=self.logger, name=kwargs.pop("name", "ThriftPyServer")
        )
        TThreadedServer.__init__(self, *args, **kwargs)

    def run(self):
        self.logger.info(self.display_name + " is started")
        self.serve()

        # We could still be waiting on worker threads to finish -
        # we need another event to tell when REALLY stopped
        self.finished.wait(timeout=self.WORKER_TIMEOUT + 1)
        self.logger.info(self.display_name + " is stopped")

    def stop(self):
        # Mark the thread as stopping
        StoppableThread.stop(self)

        # Close the socket - this will kick us out of the self.trans.accept() call with an exception
        self.trans.close()

        # Wait some amount of time for all the futures to complete
        futures_status = wait(
            self.futures, timeout=self.WORKER_TIMEOUT, return_when=ALL_COMPLETED
        )

        # If there are still workers remaining after the timeout then we remove references to them.
        # We need to do this because workers are daemons but concurrent.futures.thread adds a
        # hook to join all workers with no timeout when shutting down. So any hung worker would
        # prevent the application from shutting down.
        if futures_status.not_done:
            self.logger.warning(
                "There were still unfinished worker "
                "threads even after waiting, about to orphan them"
            )

            import concurrent.futures.thread

            self.pool._threads.clear()
            concurrent.futures.thread._threads_queues.clear()

        # Let run() know that we've made our attempt to wait
        # for all the workers and now it's time to die
        self.finished.set()

    def serve(self):
        self.trans.listen()

        while not self.stopped():
            try:
                trans_input = self.trans.accept()

                future = self.pool.submit(self.handle, trans_input)
                self.futures.add(future)
                future.add_done_callback(lambda x: self.futures.remove(x))
            except OSError as ex:
                if not self.stopped() or ex.errno != 22:
                    self.logger.exception(ex)
            except Exception as ex:
                self.logger.exception(ex)


class WrappedTProcessor(TProcessor):
    """Processor that fails gracefully if a handler method raises an unexpected exception"""

    def __init__(self, default_exception_name, default_exception_cls, *args, **kwargs):
        super(WrappedTProcessor, self).__init__(*args, **kwargs)
        self.default_exception_name = default_exception_name
        self.default_exception_cls = default_exception_cls
        self.logger = logging.getLogger(__name__)

    def handle_exception(self, e, result):
        try:
            return super(WrappedTProcessor, self).handle_exception(e, result)
        except Exception as ex:
            self.logger.exception(
                "Uncaught exception occurred during thrift execution: %s", ex
            )
            setattr(
                result,
                self.default_exception_name,
                self.default_exception_cls(
                    str(ex) or "No message was found on the raised exception."
                ),
            )


def make_server(
    service, handler, host="127.0.0.1", port=9090, client_timeout=None, cert_file=None
):
    """Factory method to create a BartenderThriftServer"""

    if cert_file:
        server_socket = TSSLServerSocket(
            host=host, port=port, client_timeout=client_timeout, certfile=cert_file
        )
    else:
        server_socket = TServerSocket(
            host=host, port=port, client_timeout=client_timeout
        )

    return BartenderThriftServer(
        WrappedTProcessor("baseEx", bg_utils.bg_thrift.BaseException, service, handler),
        server_socket,
    )



import json
import logging
from collections import Sequence

import re
import six
import urllib3
from builtins import str
from requests import Session

import bartender
from bg_utils.mongo.models import System, Choices
from brewtils.choices import parse
from brewtils.errors import ModelValidationError
from brewtils.rest.system_client import SystemClient


class RequestValidator(object):
    def __init__(self):
        self.logger = logging.getLogger(__name__)

        self._client = SystemClient(system_name=None, **bartender.config.web)

        self._session = Session()
        if not bartender.config.web.ca_verify:
            urllib3.disable_warnings()
            self._session.verify = False
        elif bartender.config.web.ca_cert:
            self._session.verify = bartender.config.web.ca_cert

    def validate_request(self, request):
        """Validation to be called before you save a request from a user

        :param request: The request to validate
        """
        self.logger.debug("Validating request")

        system = self.get_and_validate_system(request)
        command = self.get_and_validate_command_for_system(request, system)

        request.parameters = self.get_and_validate_parameters(request, command)

        return request

    def get_and_validate_system(self, request):
        """Ensure there is a system in the DB that corresponds to this Request.

        :param request: The request to validate
        :return: The system corresponding to this Request
        :raises ModelValidationError: There is no system that corresponds to this Request
        """
        system = System.find_unique(request.system, request.system_version)
        if system is None:
            raise ModelValidationError(
                "Could not find System named '%s' matching version '%s'"
                % (request.system, request.system_version)
            )

        if request.instance_name not in system.instance_names:
            raise ModelValidationError(
                "Could not find instance with name '%s' in system '%s'"
                % (request.instance_name, system.name)
            )

        self.logger.debug(
            "Found System %s-%s" % (request.system, request.instance_name)
        )
        return system

    def get_and_validate_command_for_system(self, request, system=None):
        """Ensure the System has a command with a name that matches this request.

        :param request: The request to validate
        :param system: Specifies a System to use. If None a system lookup will be attempted.
        :return: The database command
        :raises ValidationError: if the request or command is invalid
        """
        self.logger.debug("Getting and Validating Command for System")
        if system is None:
            self.logger.debug("No System was passed in")
            system = self.get_and_validate_system(request)

        if request.command is None:
            raise ModelValidationError(
                "Could not validate command because it was None."
            )

        self.logger.debug(
            "Looking through Command Names to find the Command Specified."
        )
        command_names = []
        for command in system.commands:
            if command.name == request.command:
                self.logger.debug("Found Command with name: %s" % request.command)

                if request.command_type is None:
                    request.command_type = command.command_type
                elif command.command_type != request.command_type:
                    raise ModelValidationError(
                        "Command Type for Request was %s but the command specified "
                        "the type as %s" % (request.command_type, command.command_type)
                    )

                if request.output_type is None:
                    request.output_type = command.output_type
                elif command.output_type != request.output_type:
                    raise ModelValidationError(
                        "Output Type for Request was %s but the command specified "
                        "the type as %s" % (request.output_type, command.output_type)
                    )

                return command

            command_names.append(command.name)

        raise ModelValidationError(
            "No Command with name: %s could be found. Valid Commands for %s are: %s"
            % (request.command, system.name, command_names)
        )

    def get_and_validate_parameters(
        self, request, command=None, command_parameters=None, request_parameters=None
    ):
        """Validates all request parameters

        Note: It is significant that we build the parameters as a separate object and then assign
        it. This is significant because when you are setting a value to None, mongoengine
        interprets this as "mark the item for deletion" meaning that when you save that attribute,
        it will actually get deleted. Sounds weird right? Well at least you didn't have to
        troubleshoot it.

        :param request: The Request to validate
        :param command: The the Command for this request. Will attempt to discover if None.
        :param command_parameters:  The command parameters.
        :param request_parameters: The request parameters.
        :return: The updated Parameters, ready to be saved to the database
        """
        self.logger.debug("Updating and Validating Parameters")

        if command is None:
            self.logger.debug("No Command Provided.")
            command = self.get_and_validate_command_for_system(request)

        if command_parameters is None:
            command_parameters = command.parameters

        if request_parameters is None:
            request_parameters = request.parameters

        self._validate_no_extra_request_parameter_keys(
            request_parameters, command_parameters
        )

        parameters_to_save = {}
        for command_parameter in command_parameters:
            self._validate_required_parameter_is_included_in_request(
                request, command_parameter, request_parameters
            )
            extracted_value = self._extract_parameter_value_from_request(
                request, command_parameter, request_parameters, command
            )
            self._validate_value_in_choices(request, extracted_value, command_parameter)
            self._validate_maximum(extracted_value, command_parameter)
            self._validate_minimum(extracted_value, command_parameter)
            self._validate_regex(extracted_value, command_parameter)
            parameters_to_save[command_parameter.key] = extracted_value

        self.logger.debug("Successfully Updated and Validated Parameters.")
        self.logger.debug("Parameters: %s", parameters_to_save)
        return parameters_to_save

    def _validate_value_in_choices(self, request, value, command_parameter):
        """Validate that the value(s) are valid according to the choice constraints"""
        if (
            value is not None
            and not command_parameter.optional
            and command_parameter.choices
            and command_parameter.choices.strict
        ):

            choices = command_parameter.choices

            def map_param_values(kv_pair_list):
                param_map = {}
                for param_name, param_ref in kv_pair_list:
                    if param_ref == "instance_name":
                        param_map[param_name] = request.instance_name
                    else:
                        param_map[param_name] = request.parameters[param_ref]

                return param_map

            if choices.type == "static":
                if isinstance(choices.value, list):
                    raw_allowed = choices.value
                elif isinstance(choices.value, dict):
                    key = choices.details.get("key_reference")
                    if key is None:
                        raise ModelValidationError(
                            "Unable to validate choices for parameter '%s' - Choices"
                            " with a dictionary value must specify a key_reference"
                            % command_parameter.key
                        )

                    if key == "instance_name":
                        key_reference_value = request.instance_name
                    else:
                        # Mongoengine stores None keys as 'null', so use that instead of None
                        key_reference_value = request.parameters.get(key) or "null"

                    raw_allowed = choices.value.get(key_reference_value)
                    if raw_allowed is None:
                        raise ModelValidationError(
                            "Unable to validate choices for parameter '%s' - Choices"
                            " dictionary doesn't contain an entry with key '%s'"
                            % (command_parameter.key, key_reference_value)
                        )
                else:
                    raise ModelValidationError(
                        "Unable to validate choices for parameter '%s' - Choices value"
                        " must be a list or dictionary " % command_parameter.key
                    )
            elif choices.type == "url":
                parsed_value = parse(choices.value, parse_as="url")
                query_params = map_param_values(parsed_value["args"])

                raw_allowed = json.loads(
                    self._session.get(parsed_value["address"], params=query_params).text
                )
            elif choices.type == "command":

                if isinstance(choices.value, six.string_types):
                    parsed_value = parse(choices.value, parse_as="func")
                    response = self._client.send_bg_request(
                        _command=parsed_value["name"],
                        _system_name=request.system,
                        _system_version=request.system_version,
                        _instance_name=request.instance_name,
                        **map_param_values(parsed_value["args"])
                    )
                elif isinstance(choices.value, dict):
                    parsed_value = parse(choices.value["command"], parse_as="func")
                    response = self._client.send_bg_request(
                        _command=parsed_value["name"],
                        _system_name=choices.value.get("system"),
                        _system_version=choices.value.get("version"),
                        _instance_name=choices.value.get("instance_name", "default"),
                        **map_param_values(parsed_value["args"])
                    )
                else:
                    raise ModelValidationError(
                        "Unable to validate choices for parameter '%s' - Choices value"
                        " must be a string or dictionary " % command_parameter.key
                    )

                raw_allowed = json.loads(response.output)
                if isinstance(raw_allowed, list):
                    if len(raw_allowed) < 1:
                        raise ModelValidationError(
                            "Unable to validate choices for parameter '%s' - Result "
                            "of choices query was empty list" % command_parameter.key
                        )
                else:
                    raise ModelValidationError(
                        "Unable to validate choices for parameter '%s' - Result of "
                        " choices query must be a list" % command_parameter.key
                    )
            else:
                raise ModelValidationError(
                    "Unable to validate choices for parameter '%s' - No valid type "
                    "specified (valid types are %s)"
                    % (command_parameter.key, Choices.TYPES)
                )

            # At this point raw_allowed is a list, but that list can potentially contain
            # {"value": "", "text": ""} dicts. Need to collapse those to strings
            allowed_values = []
            for allowed in raw_allowed:
                if isinstance(allowed, dict):
                    allowed_values.append(allowed["value"])
                else:
                    allowed_values.append(allowed)

            if command_parameter.multi:
                for single_value in value:
                    if single_value not in allowed_values:
                        raise ModelValidationError(
                            "Value '%s' is not a valid choice for parameter with key '%s'. "
                            "Valid choices are: %s"
                            % (single_value, command_parameter.key, allowed_values)
                        )
            else:
                if value not in allowed_values:
                    raise ModelValidationError(
                        "Value '%s' is not a valid choice for parameter with key '%s'. "
                        "Valid choices are: %s"
                        % (value, command_parameter.key, allowed_values)
                    )

    def _validate_maximum(self, value, command_parameter):
        """Validate that the value(s) are below the specified maximum"""
        if value is not None and not command_parameter.optional:
            if command_parameter.maximum:
                if isinstance(value, Sequence):
                    if len(value) > command_parameter.maximum:
                        raise ModelValidationError(
                            "Length %s is greater than the maximum allowed length (%s) "
                            "for parameter %s"
                            % (
                                len(value),
                                command_parameter.maximum,
                                command_parameter.key,
                            )
                        )
                else:
                    if value > command_parameter.maximum:
                        raise ModelValidationError(
                            "Value %s is greater than the maximum allowed value (%s) "
                            "for parameter %s"
                            % (value, command_parameter.maximum, command_parameter.key)
                        )

    def _validate_minimum(self, value, command_parameter):
        """Validate that the value(s) are above the specified minimum"""
        if value is not None and not command_parameter.optional:
            if command_parameter.minimum:
                if isinstance(value, Sequence):
                    if len(value) < command_parameter.minimum:
                        raise ModelValidationError(
                            "Length %s is less than the minimum allowed length (%s) "
                            "for parameter %s"
                            % (
                                len(value),
                                command_parameter.minimum,
                                command_parameter.key,
                            )
                        )
                else:
                    if value < command_parameter.minimum:
                        raise ModelValidationError(
                            "Value %s is less than the minimum allowed value (%s) "
                            "for parameter %s"
                            % (value, command_parameter.minimum, command_parameter.key)
                        )

    def _validate_regex(self, value, command_parameter):
        """Validate that the value matches the regex"""
        if value is not None and not command_parameter.optional:
            if command_parameter.regex:
                if not re.match(command_parameter.regex, value):
                    raise ModelValidationError(
                        "Value %s does not match regular expression %s"
                        % (value, command_parameter.regex)
                    )

    def _extract_parameter_value_from_request(
        self, request, command_parameter, request_parameters, command
    ):
        """Extracts the expected value based on the parameter in the database,
        uses the default and validates the type of the request parameter"""
        request_value = request_parameters.get(
            command_parameter.key, command_parameter.default
        )

        if request_value is None and command_parameter.nullable:
            return None

        if command_parameter.multi:
            request_values = request_value
            if not isinstance(request_values, list):
                raise ModelValidationError(
                    "%s was specified as a list, "
                    "but was not provided as such" % command_parameter.key
                )

            value_to_return = []
            for value in request_values:
                value_to_return.append(
                    self._validate_parameter_based_on_type(
                        value, command_parameter, command, request
                    )
                )
        else:
            value_to_return = self._validate_parameter_based_on_type(
                request_value, command_parameter, command, request
            )

        return value_to_return

    def _validate_required_parameter_is_included_in_request(
        self, request, command_parameter, request_parameters
    ):
        """If the parameter is required but was not provided in the request_parameters and does
        not have a default, then raise a ValidationError"""
        self.logger.debug(
            "Validating that Required Parameters are included in the request."
        )
        if not command_parameter.optional:
            if (
                command_parameter.key not in request_parameters
                and command_parameter.default is None
            ):
                raise ModelValidationError(
                    "Required key '%s' not provided in request. Parameters are: %s"
                    % (command_parameter.key, request.parameters)
                )

    def _validate_no_extra_request_parameter_keys(
        self, request_parameters, command_parameters
    ):
        """Validate that all the parameters passed in were valid keys. If there is a key specified
        that is not noted in the database, then a validation error is thrown"""
        self.logger.debug("Validating Keys")
        valid_keys = [cp.key for cp in command_parameters]
        self.logger.debug("Valid Keys are : %s" % valid_keys)
        for key in request_parameters:
            if key not in valid_keys:
                raise ModelValidationError(
                    "Unknown key '%s' provided in the parameters. Valid Keys are: %s"
                    % (key, valid_keys)
                )

    def _validate_parameter_based_on_type(self, value, parameter, command, request):
        """Validates the value passed in, ensures the type matches.
        Recursive calls for dictionaries which also have nested parameters"""

        try:
            if value is None and not parameter.nullable:
                raise ModelValidationError(
                    "There is no value for parameter '%s' "
                    "and this field is not nullable." % parameter.key
                )
            elif parameter.type.upper() == "STRING":
                if isinstance(value, six.string_types):
                    return str(value)
                else:
                    raise TypeError("Invalid value for string (%s)" % value)
            elif parameter.type.upper() == "INTEGER":
                if int(value) != float(value):
                    raise TypeError("Invalid value for integer (%s)" % value)
                return int(value)
            elif parameter.type.upper() == "FLOAT":
                return float(value)
            elif parameter.type.upper() == "ANY":
                return value
            elif parameter.type.upper() == "BOOLEAN":
                if value in [True, False]:
                    return value
                else:
                    raise TypeError("Invalid value for boolean (%s)" % value)
            elif parameter.type.upper() == "DICTIONARY":
                dict_value = dict(value)
                if parameter.parameters:
                    self.logger.debug("Found Nested Parameters.")
                    return self.get_and_validate_parameters(
                        request, command, parameter.parameters, dict_value
                    )
                return dict_value
            elif parameter.type.upper() == "DATE":
                return int(value)
            elif parameter.type.upper() == "DATETIME":
                return int(value)
            else:
                raise ModelValidationError(
                    "Unknown type for parameter. Please contact a system administrator."
                )
        except TypeError as ex:
            self.logger.exception(ex)
            raise ModelValidationError(
                "Value for key: %s is not the correct type. Should be: %s"
                % (parameter.key, parameter.type)
            )
        except ValueError as ex:
            self.logger.exception(ex)
            raise ModelValidationError(
                "Value for key: %s is not the correct type. Should be: %s"
                % (parameter.key, parameter.type)
            )



def string_contains_environment_var(string):
    """Determines whether or not there is a valid environment variable in the string"""
    parts = string.split("$")

    # If there are no $ then there is no possiblity for there to be
    # an environment variable
    if len(parts) == 1:
        return False

    for index, part in enumerate(parts):
        # If the part has no length, then the first variable is an unescaped
        # Dollarsign and we will just continue (as it could be an invalid
        # var such as: $%
        if len(part) == 0:
            continue

        # If the first character is not an alpha or an _ then this is an
        # invalid environment variable and will not be treated as such
        if not part[0].isalpha():
            continue

        if index > 0:
            # If we are not on the first element, then we check to see
            # if the previous entry ended in a "\" if it did, then that
            # means this value should have a literal dollar sign in it
            # so we just continue:
            # i.e.                      foo\$bar
            prev_part = parts[index - 1]
            if prev_part.endswith("\\"):
                continue

            # otherwise, then we've found a valid environment variable
            else:
                return True

        # If we are on the first element, and it does not end with a "\"
        # and it is not the last element in the array, then we must have
        # an environment Variable:
        # i.e.:           /path/to/foo:$PATH
        elif not part.endswith("\\") and index + 1 != len(parts):
            return True

    return False


def is_string_environment_variable(string):
    """Determines if string being sent in without the $ is a valid environmnet variable"""
    if len(string) == 0:
        return False

    return string[0].isalpha()


def get_environment_var_name_from_string(string):
    """Strips out the Variable name from a string
    :param string: e.g. $MY_PATH:$YOUR_PATH
    :return string: e.g. MY_PATH
    """
    key = ""
    for char in string:
        if char.isalpha() or char.isdigit() or char == "_":
            key += char
        else:
            return key

    return key


def expand_string_with_environment_var(string, env_copy=None):
    """Expands a String with a $VAR style var in it."""
    import os

    parts = string.split("$")
    if env_copy is None:
        env_copy = os.environ.copy()

    expanded_value = ""
    for index, part in enumerate(parts):
        if index == 0:
            if len(part) == 0:
                continue
            else:
                expanded_value += part

        elif expanded_value.endswith("\\"):
            expanded_value += "$" + part

        elif is_string_environment_variable(part):
            environment_key = get_environment_var_name_from_string(part)
            environment_value = env_copy.get(environment_key, "")
            expanded_value += part.replace(environment_key, environment_value)
        else:
            expanded_value += "$" + part

    return expanded_value



import logging

from pyrabbit2.api import Client
from pyrabbit2.http import HTTPError, NetworkError

import bartender
from bg_utils.mongo.parser import MongoParser


class PyrabbitClient(object):
    """Class that implements a connection to RabbitMQ Management HTTP API"""

    def __init__(
        self,
        host="localhost",
        port=15672,
        user="guest",
        password="guest",
        virtual_host="/",
        ssl=None,
    ):
        self.logger = logging.getLogger(__name__)

        # Pyrabbit won't infer the default virtual host ('/'). So we need to enforce it
        self._virtual_host = virtual_host or "/"

        ssl = ssl or {}
        verify = ssl.get("ca_cert", True) if ssl.get("ca_verify") else False

        # The client for doing Admin things over the HTTP API
        self._client = Client(
            "%s:%s" % (host, port),
            user,
            password,
            scheme="https" if ssl.get("enabled") else "http",
            verify=verify,
            cert=ssl.get("client_cert"),
        )

    def is_alive(self):
        try:
            return self._client.is_alive()
        except NetworkError:
            return False

    def verify_virtual_host(self):
        """Ensure the virtual host exists"""
        try:
            return self._client.get_vhost(self._virtual_host)
        except Exception:
            self.logger.error(
                "Error verifying virtual host %s, does it exist?", self._virtual_host
            )
            raise

    def ensure_admin_expiry(self):
        """Ensure that the admin queue expiration policy exists"""
        try:
            kwargs = {
                "pattern": "^admin.*",
                "definition": {"expires": bartender.config.amq.admin_queue_expiry},
                "priority": 1,
                "apply-to": "queues",
            }
            self._client.create_policy(self._virtual_host, "admin_expiry", **kwargs)
        except Exception:
            self.logger.error("Error creating admin queue expiration policy")
            raise

    def get_queue_size(self, queue_name):
        """Get the number of messages in a queue.

        :param queue_name: The name of the queue
        :return: The number of messages in the queue
        """
        self.logger.debug("Getting queue Size for: %s", queue_name)
        try:
            return self._client.get_queue(self._virtual_host, queue_name).get(
                "messages", 0
            )
        except HTTPError as ex:
            if ex.status == 404:
                self.logger.error("Queue '%s' could not be found", queue_name)
            else:
                self.logger.error("Could not connect to queue '%s'", queue_name)
            raise ex

    def clear_queue(self, queue_name):
        """Remove all messages in a queue.

        :param queue_name: The name of the queue
        :return:None
        """
        self.logger.info("Clearing Queue: %s", queue_name)
        queue_dictionary = self._client.get_queue(self._virtual_host, queue_name)
        number_of_messages = queue_dictionary.get("messages_ready", 0)

        while number_of_messages > 0:
            self.logger.debug("Getting the Next Message")
            messages = self._client.get_messages(
                self._virtual_host, queue_name, count=1, requeue=False
            )
            if messages and len(messages) > 0:
                message = messages[0]
                try:
                    request = MongoParser.parse_request(
                        message["payload"], from_string=True
                    )
                    self.logger.debug("Canceling Request: %s", request.id)
                    bartender.bv_client.update_request(request.id, status="CANCELED")
                except Exception as ex:
                    self.logger.error("Error removing message:")
                    self.logger.exception(ex)
            else:
                self.logger.debug(
                    "Race condition: The while loop thought there were "
                    "more messages to ingest but no more messages could "
                    "be received."
                )
                break

            number_of_messages -= 1

    def delete_queue(self, queue_name):
        """Actually remove a queue.

        :param queue_name: The name of the queue
        :return:
        """
        self._client.delete_queue(self._virtual_host, queue_name)

    def destroy_queue(self, queue_name, force_disconnect=False):
        """Remove all remnants of a queue.

        Ignores exceptions and ensures all aspects of the queue are deleted.

        :param queue_name: The queue name
        :param force_disconnect: Attempt to forcefully disconnect consumers of this queue
        :return:
        """
        if queue_name is None:
            return

        if force_disconnect:
            try:
                self.disconnect_consumers(queue_name)
            except HTTPError as ex:
                if ex.status != 404:
                    self.logger.exception(ex)
            except Exception as ex:
                self.logger.exception(ex)

        try:
            self.clear_queue(queue_name)
        except HTTPError as ex:
            if ex.status != 404:
                self.logger.exception(ex)
        except Exception as ex:
            self.logger.exception(ex)

        try:
            self.delete_queue(queue_name)
        except HTTPError as ex:
            if ex.status != 404:
                self.logger.exception(ex)
        except Exception as ex:
            self.logger.exception(ex)

    def disconnect_consumers(self, queue_name):
        # If there are no channels, then there is nothing to do
        channels = self._client.get_channels() or []
        for channel in channels:

            # If the channel is already gone, just return an empty response
            channel_details = self._client.get_channel(channel["name"]) or {
                "consumer_details": []
            }

            for consumer_details in channel_details["consumer_details"]:
                if queue_name == consumer_details["queue"]["name"]:
                    self._client.delete_connection(
                        consumer_details["channel_details"]["connection_name"]
                    )



import logging

from brewtils.stoppable_thread import StoppableThread


class LocalPluginMonitor(StoppableThread):
    """Object to constantly monitor that plugins are alive and working.

    When one is down and out, it will attempt to restart that plugin.
    """

    def __init__(self, plugin_manager, registry):
        self.logger = logging.getLogger(__name__)
        self.display_name = "Local Plugin Monitor"
        self.plugin_manager = plugin_manager
        self.registry = registry

        super(LocalPluginMonitor, self).__init__(
            logger=self.logger, name="LocalPluginMonitor"
        )

    def run(self):
        self.logger.info(self.display_name + " is started")

        while not self.wait(1):
            self.monitor()

        self.logger.info(self.display_name + " is stopped")

    def monitor(self):
        """Make sure plugins stay alive.

        Iterate through all plugins, testing them one at a time.
        If any of them are dead restart them, otherwise just keep chugging along.
        """
        for plugin in self.registry.get_all_plugins():
            if self.stopped():
                break

            if (
                plugin.process
                and plugin.process.poll() is not None
                and not plugin.stopped()
            ):
                if plugin.status == "RUNNING":
                    self.logger.warning(
                        "It looks like plugin %s has " "unexpectedly stopped running.",
                        plugin.unique_name,
                    )
                    self.logger.warning(
                        "If this is happening often, you "
                        "need to talk to the plugin developer."
                    )
                    self.logger.warning("Restarting plugin: %s", plugin.unique_name)

                    plugin.status = "DEAD"
                    self.plugin_manager.restart_plugin(plugin)
                elif plugin.status == "STARTING":
                    self.logger.warning(
                        "It looks like plugin %s has " "failed to start.",
                        plugin.unique_name,
                    )
                    self.logger.warning(
                        "Marking plugin %s as dead.", plugin.unique_name
                    )
                    plugin.status = "DEAD"



import logging

import bartender
from bartender.errors import PluginStartupError
from bartender.local_plugins.plugin_runner import LocalPluginRunner
from bg_utils.mongo.models import System


class LocalPluginsManager(object):
    """LocalPluginsManager that is capable of stopping/starting and restarting plugins"""

    def __init__(self, loader, validator, registry, clients):
        self.logger = logging.getLogger(__name__)
        self.loader = loader
        self.validator = validator
        self.registry = registry
        self.clients = clients

    def start_plugin(self, plugin):
        """Start a specific plugin.

        If a plugin cannot Be found (i.e. it was not loaded) then it will not do anything
        If the plugin is already running, it will do nothing.
        If a plugin instance has not started after the PLUGIN_STARTUP_TIMEOUT it will mark that
        instance as stopped

        :param plugin: The plugin to start
        :return: True if any plugin instances successfully started. False if the plugin does
            not exist or all instances failed to start
        """
        self.logger.info("Starting plugin %s", plugin.unique_name)

        if plugin.status in ["RUNNING", "STARTING"]:
            self.logger.info("Plugin %s is already running.", plugin.unique_name)
            return True

        if plugin.status == "INITIALIZING":
            new_plugin = plugin
        elif plugin.status in ["DEAD", "STOPPED"]:
            new_plugin = LocalPluginRunner(
                plugin.entry_point,
                plugin.system,
                plugin.instance_name,
                plugin.path_to_plugin,
                plugin.web_host,
                plugin.web_port,
                plugin.ssl_enabled,
                plugin_args=plugin.plugin_args,
                environment=plugin.environment,
                requirements=plugin.requirements,
                plugin_log_directory=plugin.plugin_log_directory,
                url_prefix=plugin.url_prefix,
                ca_verify=plugin.ca_verify,
                ca_cert=plugin.ca_cert,
                username=plugin.username,
                password=plugin.password,
            )
            self.registry.remove(plugin.unique_name)
            self.registry.register_plugin(new_plugin)
        else:
            raise PluginStartupError("Plugin in an invalid state (%s)" % plugin.status)

        new_plugin.status = "STARTING"
        new_plugin.start()

        return True

    def stop_plugin(self, plugin):
        """Stops a Plugin

        :param plugin The plugin to stop.
        :return: None
        """
        self.logger.info("Stopping plugin %s", plugin.unique_name)

        # Need to mark the plugin as dead if it doesn't shut down cleanly
        clean_shutdown = True

        try:
            if plugin.status in ["DEAD", "STOPPED", "STOPPING"]:
                self.logger.info("Plugin %s was already stopped", plugin.unique_name)
                return
            elif plugin.status == "UNKNOWN":
                self.logger.warning(
                    "Couldn't determine status of plugin %s, "
                    "still attempting to stop",
                    plugin.unique_name,
                )
            else:
                plugin.status = "STOPPING"

            # Plugin must be marked as stopped before sending shutdown message
            plugin.stop()

            # Send a stop request. This initiates a graceful shutdown on the plugin side.
            self.clients["pika"].stop(
                system=plugin.system.name,
                version=plugin.system.version,
                instance=plugin.instance_name,
            )

            # Now just wait for the plugin thread to die
            self.logger.info("Waiting for plugin %s to stop...", plugin.unique_name)
            plugin.join(bartender.config.plugin.local.timeout.shutdown)

        except Exception as ex:
            clean_shutdown = False
            self.logger.error(
                "Plugin %s raised exception while shutting down:", plugin.unique_name
            )
            self.logger.exception(ex)

        finally:
            if plugin.is_alive():
                self.logger.error(
                    "Plugin %s didn't terminate, about to kill", plugin.unique_name
                )
                plugin.kill()
                clean_shutdown = False

        # Local plugins will set their status to STOPPED in their stop handler
        if not clean_shutdown:
            self.logger.warning(
                "Plugin %s did not shutdown cleanly, " "marking as DEAD",
                plugin.unique_name,
            )
            plugin.status = "DEAD"

    def restart_plugin(self, plugin):
        self.stop_plugin(plugin)
        self.start_plugin(plugin)

    def reload_system(self, system_name, system_version):
        """Reload a specific system

        :param system_name: The name of the system to reload
        :param system_version: The version of the system to reload
        :return: None
        """
        plugins = self.registry.get_plugins_by_system(system_name, system_version)
        if len(plugins) < 1:
            message = "Could not reload system %s-%s: not found in the registry" % (
                system_name,
                system_version,
            )
            self.logger.error(message)
            raise Exception(message)

        path_to_plugin = plugins[0].path_to_plugin

        # Verify the new configuration is valid before we remove the
        # current plugins from the registry
        if not self.validator.validate_plugin(path_to_plugin):
            message = (
                "Could not reload system %s-%s: new configuration is not valid"
                % (system_name, system_version)
            )
            self.logger.warning(message)
            raise Exception(message)

        for plugin in plugins:
            if plugin.status == "RUNNING":
                message = "Could not reload system %s-%s: running instances" % (
                    system_name,
                    system_version,
                )
                self.logger.warning(message)
                raise Exception(message)

        for plugin in plugins:
            self.registry.remove(plugin.unique_name)

        self.loader.load_plugin(path_to_plugin)

    def start_all_plugins(self):
        """Attempts to start all plugins in the registry."""
        self.logger.info("Starting all plugins")
        self._start_multiple_plugins(self.registry.get_all_plugins())

    def _start_multiple_plugins(self, plugins_to_start):
        """Starts multiple plugins, respecting required plugin dependencies.

        It will consider requirements of other plugins. If the plugin requires others, then it
        will be skipped until its requirements are loaded.

        Failure is a little tricky.

        :param plugins_to_start: List of plugin names to be started
        :return: None
        """
        start_list = list(plugins_to_start)

        system_names = self._get_all_system_names()
        failed_system_names = self._get_failed_system_names()
        started_plugin_names = self._get_running_system_names()

        while len(start_list) != 0:
            plugin = start_list.pop()
            attempt_to_start = True

            self.logger.debug("Checking plugin %s's requirements.", plugin.unique_name)
            for required_plugin_name in plugin.requirements:
                if required_plugin_name not in system_names:
                    self.logger.warning(
                        "Plugin %s lists system %s as a required system, "
                        "but that system is not available.",
                        plugin.unique_name,
                        required_plugin_name,
                    )
                    self._mark_as_failed(plugin)
                    failed_system_names.append(plugin.system.name)
                    attempt_to_start = False
                    break

                elif required_plugin_name in failed_system_names:
                    self.logger.warning(
                        "Plugin %s lists plugin %s as a required plugin, "
                        "but plugin %s failed to start,"
                        " thus plugin %s cannot start.",
                        plugin.unique_name,
                        required_plugin_name,
                        required_plugin_name,
                        plugin.unique_name,
                    )
                    self._mark_as_failed(plugin)
                    failed_system_names.append(plugin.system.name)
                    attempt_to_start = False
                    break

                elif required_plugin_name not in started_plugin_names:
                    self.logger.debug(
                        "Skipping Starting Plugin %s because its "
                        "requirements have yet to be started.",
                        plugin.unique_name,
                    )
                    start_list.insert(0, plugin)
                    attempt_to_start = False
                    break

            if attempt_to_start:
                if self.start_plugin(plugin):
                    started_plugin_names.append(plugin.system.name)
                else:
                    failed_system_names.append(plugin.system.name)

        self.logger.info("Finished starting plugins.")

    @staticmethod
    def _get_all_system_names():
        return [system.name for system in System.objects()]

    @staticmethod
    def _get_running_system_names():
        running_system_names = []
        for system in System.objects():
            if all([instance.status == "RUNNING" for instance in system.instances]):
                running_system_names.append(system.name)

        return running_system_names

    @staticmethod
    def _get_failed_system_names():
        failed_system_names = []
        for system in System.objects():
            if all([instance.status == "DEAD" for instance in system.instances]):
                failed_system_names.append(system.name)

        return failed_system_names

    @staticmethod
    def _mark_as_failed(plugin):
        system = System.find_unique(plugin.system.name, plugin.system.version)
        for instance in system.instances:
            if instance.name == plugin.instance_name:
                instance.status = "DEAD"
        system.deep_save()

    def stop_all_plugins(self):
        """Attempt to stop all plugins."""
        self.logger.info("Stopping all plugins")

        for plugin in self.registry.get_all_plugins():
            try:
                self.stop_plugin(plugin)
            except Exception as ex:
                self.logger.error("Error stopping plugin %s", plugin.unique_name)
                self.logger.exception(ex)

    def scan_plugin_path(self):
        """Scans the default plugin directory for new plugins.

        Will also start any new plugins found.

        :return: None
        """
        scanned_plugins_paths = set(self.loader.scan_plugin_path())
        existing_plugin_paths = set(
            [plugin.path_to_plugin for plugin in self.registry.get_all_plugins()]
        )

        new_plugins = []
        for plugin_path in scanned_plugins_paths.difference(existing_plugin_paths):
            try:
                loaded_plugins = self.loader.load_plugin(plugin_path)

                if not loaded_plugins:
                    raise Exception("Couldn't load plugin at %s" % plugin_path)

                for plugin in loaded_plugins:
                    new_plugins.append(plugin)
            except Exception as ex:
                self.logger.error(
                    "Error while attempting to load plugin at %s", plugin_path
                )
                self.logger.exception(ex)

        self._start_multiple_plugins(new_plugins)

    def pause_plugin(self, unique_name):
        """Pause a plugin. Not Used yet."""
        plugin = self.registry.get_plugin(unique_name)

        if plugin is None:
            self.logger.warning("Plugin %s is not loaded.", unique_name)
            return

        if plugin.status == "RUNNING":
            plugin.status = "PAUSED"

    def unpause_plugin(self, unique_name):
        """Unpause a plugin. Not Used yet."""
        plugin = self.registry.get_plugin(unique_name)

        if plugin is None:
            self.logger.warning("Plugin %s is not loaded.", unique_name)
            return

        if plugin.status == "PAUSED":
            plugin.status = "RUNNING"



import logging
from datetime import datetime, timedelta

from mongoengine import Q

from brewtils.stoppable_thread import StoppableThread


class MongoPruner(StoppableThread):
    def __init__(self, tasks=None, run_every=timedelta(minutes=15)):
        self.logger = logging.getLogger(__name__)
        self.display_name = "Mongo Pruner"
        self._run_every = run_every.total_seconds()
        self._tasks = tasks or []

        super(MongoPruner, self).__init__(logger=self.logger, name="Remover")

    def add_task(
        self, collection=None, field=None, delete_after=None, additional_query=None
    ):
        self._tasks.append(
            {
                "collection": collection,
                "field": field,
                "delete_after": delete_after,
                "additional_query": additional_query,
            }
        )

    def run(self):
        self.logger.info(self.display_name + " is started")

        while not self.wait(self._run_every):
            current_time = datetime.utcnow()

            for task in self._tasks:
                delete_older_than = current_time - task["delete_after"]

                query = Q(**{task["field"] + "__lt": delete_older_than})
                if task.get("additional_query", None):
                    query = query & task["additional_query"]

                self.logger.debug(
                    "Removing %ss older than %s"
                    % (task["collection"].__name__, str(delete_older_than))
                )
                task["collection"].objects(query).no_cache().delete()

        self.logger.info(self.display_name + " is stopped")



SPECIFICATION = {
    "configuration": {
        "type": "dict",
        "bootstrap": True,
        "items": {
            "file": {
                "type": "str",
                "description": "Path to configuration file to use",
                "required": False,
                "cli_short_name": "c",
                "bootstrap": True,
                "previous_names": ["config"],
                "alt_env_names": ["CONFIG"],
            },
            "type": {
                "type": "str",
                "description": "Configuration file type",
                "required": False,
                "cli_short_name": "t",
                "bootstrap": True,
                "choices": ["json", "yaml"],
            },
        },
    },
    "publish_hostname": {
        "type": "str",
        "default": "localhost",
        "description": "Publicly accessible hostname for plugins to connect to",
        "previous_names": ["amq_publish_host"],
        "alt_env_names": ["AMQ_PUBLISH_HOST"],
    },
    "amq": {
        "type": "dict",
        "items": {
            "host": {
                "type": "str",
                "default": "localhost",
                "description": "Hostname of AMQ to use",
                "previous_names": ["amq_host"],
            },
            "admin_queue_expiry": {
                "type": "int",
                "default": 3600000,  # One hour
                "description": "Time before unused admin queues are removed",
            },
            "heartbeat_interval": {
                "type": "int",
                "default": 3600,
                "description": "Heartbeat interval for AMQ",
                "previous_names": ["amq_heartbeat_interval"],
            },
            "blocked_connection_timeout": {
                "type": "int",
                "default": 5,
                "description": "Time to wait for a blocked connection to be unblocked",
            },
            "connection_attempts": {
                "type": "int",
                "default": 3,
                "description": "Number of retries to connect to AMQ",
                "previous_names": ["amq_connection_attempts"],
            },
            "exchange": {
                "type": "str",
                "default": "beer_garden",
                "description": "Exchange name to use for AMQ",
                "previous_names": ["amq_exchange"],
            },
            "virtual_host": {
                "type": "str",
                "default": "/",
                "description": "Virtual host to use for AMQ",
                "previous_names": ["amq_virtual_host"],
            },
            "connections": {
                "type": "dict",
                "items": {
                    "admin": {
                        "type": "dict",
                        "items": {
                            "port": {
                                "type": "int",
                                "default": 15672,
                                "description": "Port of the AMQ Admin host",
                                "previous_names": ["amq_admin_port"],
                                "alt_env_names": ["AMQ_ADMIN_PORT"],
                            },
                            "user": {
                                "type": "str",
                                "default": "guest",
                                "description": "Username to login to the AMQ admin",
                                "previous_names": ["amq_admin_user"],
                                "alt_env_names": ["AMQ_ADMIN_USER"],
                            },
                            "password": {
                                "type": "str",
                                "default": "guest",
                                "description": "Password to login to the AMQ admin",
                                "previous_names": [
                                    "amq_admin_password",
                                    "amq_admin_pw",
                                ],
                                "alt_env_names": ["AMQ_ADMIN_PASSWORD", "AMQ_ADMIN_PW"],
                            },
                            "ssl": {
                                "type": "dict",
                                "items": {
                                    "enabled": {
                                        "type": "bool",
                                        "default": False,
                                        "description": "Should the connection use SSL",
                                    },
                                    "ca_cert": {
                                        "type": "str",
                                        "description": "Path to CA certificate file to use",
                                        "required": False,
                                    },
                                    "ca_verify": {
                                        "type": "bool",
                                        "default": True,
                                        "description": "Verify external certificates",
                                        "required": False,
                                    },
                                    "client_cert": {
                                        "type": "str",
                                        "description": "Path to client combined key / certificate",
                                        "required": False,
                                    },
                                },
                            },
                        },
                    },
                    "message": {
                        "type": "dict",
                        "items": {
                            "port": {
                                "type": "int",
                                "default": 5672,
                                "description": "Port of the AMQ host",
                                "previous_names": ["amq_port"],
                                "alt_env_names": ["AMQ_PORT"],
                            },
                            "password": {
                                "type": "str",
                                "default": "guest",
                                "description": "Password to login to the AMQ host",
                                "previous_names": ["amq_password"],
                                "alt_env_names": ["AMQ_PASSWORD"],
                            },
                            "user": {
                                "type": "str",
                                "default": "guest",
                                "description": "Username to login to the AMQ host",
                                "previous_names": ["amq_user"],
                                "alt_env_names": ["AMQ_USER"],
                            },
                            "ssl": {
                                "type": "dict",
                                "items": {
                                    "enabled": {
                                        "type": "bool",
                                        "default": False,
                                        "description": "Should the connection use SSL",
                                    },
                                    "ca_cert": {
                                        "type": "str",
                                        "description": "Path to CA certificate file to use",
                                        "required": False,
                                    },
                                    "ca_verify": {
                                        "type": "bool",
                                        "default": True,
                                        "description": "Verify external certificates",
                                        "required": False,
                                    },
                                    "client_cert": {
                                        "type": "str",
                                        "description": "Path to client combined key / certificate",
                                        "required": False,
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    },
    "db": {
        "type": "dict",
        "items": {
            "name": {
                "type": "str",
                "default": "beer_garden",
                "description": "Name of the database to use",
                "previous_names": ["db_name"],
            },
            "connection": {
                "type": "dict",
                "items": {
                    "host": {
                        "type": "str",
                        "default": "localhost",
                        "description": "Hostname/IP of the database server",
                        "previous_names": ["db_host"],
                        "alt_env_names": ["DB_HOST"],
                    },
                    "password": {
                        "type": "str",
                        "default": None,
                        "required": False,
                        "description": "Password to connect to the database",
                        "previous_names": ["db_password"],
                        "alt_env_names": ["DB_PASSWORD"],
                    },
                    "port": {
                        "type": "int",
                        "default": 27017,
                        "description": "Port of the database server",
                        "previous_names": ["db_port"],
                        "alt_env_names": ["DB_PORT"],
                    },
                    "username": {
                        "type": "str",
                        "default": None,
                        "required": False,
                        "description": "Username to connect to the database",
                        "previous_names": ["db_username"],
                        "alt_env_names": ["DB_USERNAME"],
                    },
                },
            },
            "ttl": {
                "type": "dict",
                "items": {
                    "event": {
                        "type": "int",
                        "default": 15,
                        "description": "Number of minutes to wait before deleting "
                        "events (negative number for never)",
                        "previous_names": ["event_mongo_ttl"],
                        "alt_env_names": ["EVENT_MONGO_TTL"],
                    },
                    "action": {
                        "type": "int",
                        "default": -1,
                        "description": "Number of minutes to wait before deleting "
                        "ACTION requests (negative number for never)",
                        "previous_names": ["action_request_ttl"],
                        "alt_env_names": ["ACTION_REQUEST_TTL"],
                    },
                    "info": {
                        "type": "int",
                        "default": 15,
                        "description": "Number of minutes to wait before deleting INFO request",
                        "previous_names": ["info_request_ttl"],
                        "alt_env_names": ["INFO_REQUEST_TTL"],
                    },
                },
            },
        },
    },
    "log": {
        "type": "dict",
        "items": {
            "config_file": {
                "type": "str",
                "description": "Path to a logging config file.",
                "required": False,
                "cli_short_name": "l",
                "previous_names": ["log_config"],
                "alt_env_names": ["LOG_CONFIG"],
            },
            "file": {
                "type": "str",
                "description": "File you would like the application to log to",
                "required": False,
                "previous_names": ["log_file"],
            },
            "level": {
                "type": "str",
                "description": "Log level for the application",
                "default": "INFO",
                "choices": ["DEBUG", "INFO", "WARN", "WARNING", "ERROR", "CRITICAL"],
                "previous_names": ["log_level"],
            },
        },
    },
    "web": {
        "type": "dict",
        "items": {
            "ca_cert": {
                "type": "str",
                "description": "Path to CA certificate file to use",
                "required": False,
                "previous_names": ["ca_cert"],
                "alt_env_names": ["CA_CERT"],
            },
            "ca_verify": {
                "type": "bool",
                "default": True,
                "description": "Verify external certificates",
                "required": False,
                "previous_names": ["ca_verify"],
                "alt_env_names": ["CA_VERIFY"],
            },
            "host": {
                "type": "str",
                "default": "localhost",
                "description": "Hostname of the API server",
                "previous_names": ["web_host"],
            },
            "port": {
                "type": "int",
                "default": 2337,
                "description": "Port of the API server",
                "previous_names": ["web_port"],
            },
            "ssl_enabled": {
                "type": "bool",
                "default": False,
                "description": "Is the API server using SSL",
                "previous_names": ["ssl_enabled"],
                "alt_env_names": ["SSL_ENABLED"],
                "cli_separator": "_",
            },
            "username": {
                "type": "str",
                "description": "Username that Bartender will use for "
                "authentication (needs bg-all role)",
                "required": False,
            },
            "password": {
                "type": "str",
                "description": "Password that Bartender will use for "
                "authentication (needs bg-all role)",
                "required": False,
            },
            "url_prefix": {
                "type": "str",
                "default": None,
                "description": "URL prefix of the API server",
                "required": False,
                "previous_names": ["url_prefix"],
                "alt_env_names": ["URL_PREFIX"],
            },
        },
    },
    "thrift": {
        "type": "dict",
        "items": {
            "max_workers": {
                "type": "int",
                "default": 25,
                "description": "Maximum number of threads for handling incoming thrift calls",
                "previous_names": ["max_thrift_workers"],
                "alt_env_names": ["MAX_THRIFT_WORKERS"],
            },
            "host": {
                "type": "str",
                "default": "0.0.0.0",
                "description": "Host to bind the thrift server to",
                "previous_names": ["thrift_host"],
            },
            "port": {
                "type": "int",
                "default": 9090,
                "description": "Port to bind the thrift server to",
                "previous_names": ["thrift_port"],
            },
        },
    },
    "plugin": {
        "type": "dict",
        "items": {
            "status_heartbeat": {
                "type": "int",
                "default": 10,
                "description": "Amount of time between status messages",
                "previous_names": ["plugin_status_heartbeat"],
            },
            "status_timeout": {
                "type": "int",
                "default": 30,
                "description": "Amount of time to wait before marking a plugin as unresponsive",
                "previous_names": ["plugin_status_timeout "],
            },
            "local": {
                "type": "dict",
                "items": {
                    "auth": {
                        "type": "dict",
                        "items": {
                            "username": {
                                "type": "str",
                                "description": "Username that local plugins will use for "
                                "authentication (needs bg-plugin role)",
                                "required": False,
                            },
                            "password": {
                                "type": "str",
                                "description": "Password that local plugins will use for "
                                "authentication (needs bg-plugin role)",
                                "required": False,
                            },
                        },
                    },
                    "directory": {
                        "type": "str",
                        "description": "Directory where local plugins are located",
                        "required": False,
                        "previous_names": ["plugins_directory", "plugin_directory"],
                        "alt_env_names": ["PLUGINS_DIRECTORY", "BG_PLUGIN_DIRECTORY"],
                    },
                    "host_env_vars": {
                        "type": "list",
                        "items": {"env_var": {"type": "str"}},
                        "default": [],
                        "description": "Host environment variables that will be "
                        "propagated to local plugin processes",
                    },
                    "log_directory": {
                        "type": "str",
                        "description": "Directory where local plugin logs should go",
                        "required": False,
                        "previous_names": ["plugin_log_directory"],
                        "alt_env_names": ["PLUGIN_LOG_DIRECTORY"],
                    },
                    "timeout": {
                        "type": "dict",
                        "items": {
                            "shutdown": {
                                "type": "int",
                                "default": 10,
                                "description": "Seconds to wait for a plugin to stop gracefully",
                                "previous_names": ["plugin_shutdown_timeout"],
                                "alt_env_names": ["PLUGIN_SHUTDOWN_TIMEOUT"],
                            },
                            "startup": {
                                "type": "int",
                                "default": 5,
                                "description": "Seconds to wait for a plugin to start",
                                "previous_names": ["plugin_startup_timeout"],
                                "alt_env_names": ["PLUGIN_STARTUP_TIMEOUT"],
                            },
                        },
                    },
                },
            },
        },
    },
}


def get_default_logging_config(level, filename):
    if filename:
        bartender_handler = {
            "class": "logging.handlers.RotatingFileHandler",
            "level": level,
            "formatter": "simple",
            "filename": filename,
            "maxBytes": 10485760,
            "backupCount": 20,
            "encoding": "utf8",
        }
    else:
        bartender_handler = {
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "level": level,
            "stream": "ext://sys.stdout",
        }

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"}
        },
        "handlers": {"bartender": bartender_handler},
        "loggers": {
            "pika": {"level": "ERROR"},
            "requests.packages.urllib3.connectionpool": {"level": "WARN"},
        },
        "root": {"level": level, "handlers": ["bartender"]},
    }



#!/usr/bin/env python

import signal
import sys
from argparse import ArgumentParser
from functools import partial

from yapconf import YapconfSpec

import bartender
import bg_utils
from bartender import progressive_backoff
from bartender.specification import SPECIFICATION, get_default_logging_config
from bg_utils.mongo import setup_database


def signal_handler(signal_number, stack_frame):
    bartender.logger.info("Last call! Looks like we gotta shut down.")
    bartender.application.stop()

    bartender.logger.info(
        "Closing time! You don't have to go home, but you can't stay here."
    )
    if bartender.application.is_alive():
        bartender.application.join()

    bartender.logger.info("Looks like the Application is shut down. Have a good night!")


def generate_logging_config():
    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")
    bg_utils.generate_logging_config_file(
        spec, get_default_logging_config, sys.argv[1:]
    )


def generate_config():
    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")
    bg_utils.generate_config_file(spec, sys.argv[1:])


def migrate_config():
    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")
    bg_utils.update_config_file(spec, sys.argv[1:])


def main():
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")
    parser = ArgumentParser()
    spec.add_arguments(parser)
    args = parser.parse_args(sys.argv[1:])

    bartender.setup_bartender(spec=spec, cli_args=vars(args))

    # Ensure we have a brew-view connection
    progressive_backoff(
        partial(bartender.bv_client.can_connect, timeout=5),
        bartender.application,
        "Unable to connect to brew-view, is it started?",
    )

    # Ensure we have a mongo connection
    progressive_backoff(
        partial(setup_database, bartender.config),
        bartender.application,
        "Unable to connect to mongo, is it started?",
    )

    # Ensure we have message queue connections
    progressive_backoff(
        bartender.application.clients["pika"].is_alive,
        bartender.application,
        "Unable to connect to rabbitmq, is it started?",
    )
    progressive_backoff(
        bartender.application.clients["pyrabbit"].is_alive,
        bartender.application,
        "Unable to connect to rabbitmq admin interface. "
        "Is the management plugin enabled?",
    )

    # Since we wait for RabbitMQ and brew-view we could already be shutting down
    # In that case we don't want to start
    if not bartender.application.stopped():
        # Make sure that the bartender user has admin permissions
        bartender.ensure_admin()

        bartender.logger.info("Hi, what can I get you to drink?")
        bartender.application.start()

        bartender.logger.info("Let me know if you need anything else!")

        # You may be wondering why we don't just call bartender.application.join() or .wait().
        # Well, you're in luck because I'm going to tell you why. Either of these methods
        # cause the main python thread to lock out our signal handler, which means we cannot
        # shut down gracefully in some circumstances. So instead we simply use pause() to wait
        # for a signal to be sent to us. If you choose to change this please test thoroughly
        # when deployed via system packages (apt/yum) as well as python packages and docker.
        # Thanks!
        signal.pause()

    bartender.logger.info("Don't forget to drive safe!")


if __name__ == "__main__":
    main()



import logging


class LocalPluginRegistry(object):
    """Plugin Registry that is responsible for keeping track of plugins and their status"""

    def __init__(self):
        self._registry = []
        self.logger = logging.getLogger(__name__)

    def get_all_plugins(self):
        """Return a safe copy of all plugins"""
        return [p for p in self._registry]

    def get_unique_plugin_names(self):
        """Return a unique set of plugin names"""
        return set([p.system.name for p in self._registry])

    def get_plugin(self, unique_name):
        """Returns an actual plugin

        :param unique_name: The unique name of the plugin, in the form of name[instance]-version
        :return: The plugin
        """
        for plugin in self._registry:
            if plugin.unique_name == unique_name:
                return plugin

    def get_plugins_by_system(self, system_name, system_version):
        """Returns a list of plugins with the given system name and version

        :param system_name: The system name
        :param system_version: The system version
        :return: List of plugins
        """
        return [
            p
            for p in self._registry
            if p.system.name == system_name and p.system.version == system_version
        ]

    def remove(self, unique_name):
        """Removes a plugin from the registry

        If a plugin is removed from the registry, it must go through the re-loading
        process via the PluginLoader. Only used when there is no need for the plugin
        anymore (system shutdown or plugin is failing to start)

        :param unique_name: The unique name of the plugin, in the form of name[instance]-version
        :return: None
        """
        for index, plugin in enumerate(self._registry):
            if plugin.unique_name == unique_name:
                self.logger.info(
                    "Removing Plugin %s from the registry.", plugin.unique_name
                )
                self._registry.pop(index)
                return

    def register_plugin(self, plugin):
        """Insert a plugin into the registry

        :param plugin: The plugin to add
        :return: None
        """
        self.logger.debug("Registering plugin %s", plugin.unique_name)

        registered_plugin = self.get_plugin(plugin.unique_name)
        if registered_plugin is None:
            self.logger.info(
                "Registering Plugin %s in the registry.", plugin.unique_name
            )
            self._registry.append(plugin)
        else:
            self.logger.info(
                "Plugin %s is already in the registry.", registered_plugin.unique_name
            )

    def plugin_exists(self, plugin_name, plugin_version):
        """Query if a specific plugin exists in the registry

        :param plugin_name: The plugin name
        :param plugin_version: The plugin version
        :return:
        """
        for plugin in self._registry:
            if (
                plugin.system.name == plugin_name
                and plugin.system.version == plugin_version
            ):
                return True
        return False

    @staticmethod
    def get_unique_name(plugin_name, plugin_version, instance_name):
        return "%s[%s]-%s" % (plugin_name, instance_name, plugin_version)



from __future__ import division

import logging
from datetime import timedelta
from functools import partial

import time
from mongoengine import Q
from requests.exceptions import RequestException

import bartender
import bg_utils
from bartender.local_plugins.loader import LocalPluginLoader
from bartender.local_plugins.manager import LocalPluginsManager
from bartender.local_plugins.monitor import LocalPluginMonitor
from bartender.local_plugins.registry import LocalPluginRegistry
from bartender.local_plugins.validator import LocalPluginValidator
from bartender.mongo_pruner import MongoPruner
from bartender.monitor import PluginStatusMonitor
from bartender.pika import PikaClient
from bartender.pyrabbit import PyrabbitClient
from bartender.request_validator import RequestValidator
from bartender.thrift.handler import BartenderHandler
from bartender.thrift.server import make_server
from bg_utils.mongo.models import Event, Request
from brewtils.models import Events
from brewtils.stoppable_thread import StoppableThread


class BartenderApp(StoppableThread):
    """Main Application that Runs the Beergarden Backend."""

    def __init__(self):
        self.logger = logging.getLogger(__name__)

        self.request_validator = RequestValidator()
        self.plugin_registry = LocalPluginRegistry()
        self.plugin_validator = LocalPluginValidator()

        self.plugin_loader = LocalPluginLoader(
            validator=self.plugin_validator, registry=self.plugin_registry
        )

        self.clients = {
            "pika": PikaClient(
                host=bartender.config.amq.host,
                port=bartender.config.amq.connections.message.port,
                ssl=bartender.config.amq.connections.message.ssl,
                user=bartender.config.amq.connections.admin.user,
                password=bartender.config.amq.connections.admin.password,
                virtual_host=bartender.config.amq.virtual_host,
                connection_attempts=bartender.config.amq.connection_attempts,
                blocked_connection_timeout=bartender.config.amq.blocked_connection_timeout,
                exchange=bartender.config.amq.exchange,
            ),
            "pyrabbit": PyrabbitClient(
                host=bartender.config.amq.host,
                virtual_host=bartender.config.amq.virtual_host,
                **bartender.config.amq.connections.admin
            ),
            "public": PikaClient(
                host=bartender.config.publish_hostname,
                virtual_host=bartender.config.amq.virtual_host,
                **bartender.config.amq.connections.message
            ),
        }

        self.plugin_manager = LocalPluginsManager(
            loader=self.plugin_loader,
            validator=self.plugin_validator,
            registry=self.plugin_registry,
            clients=self.clients,
        )

        self.handler = BartenderHandler(
            registry=self.plugin_registry,
            clients=self.clients,
            plugin_manager=self.plugin_manager,
            request_validator=self.request_validator,
        )

        self.helper_threads = [
            HelperThread(
                make_server,
                service=bg_utils.bg_thrift.BartenderBackend,
                handler=self.handler,
                host=bartender.config.thrift.host,
                port=bartender.config.thrift.port,
            ),
            HelperThread(
                LocalPluginMonitor,
                plugin_manager=self.plugin_manager,
                registry=self.plugin_registry,
            ),
            HelperThread(
                PluginStatusMonitor,
                self.clients,
                timeout_seconds=bartender.config.plugin.status_timeout,
                heartbeat_interval=bartender.config.plugin.status_heartbeat,
            ),
        ]

        # Only want to run the MongoPruner if it would do anything
        tasks, run_every = self._setup_pruning_tasks()
        if run_every:
            self.helper_threads.append(
                HelperThread(
                    MongoPruner, tasks=tasks, run_every=timedelta(minutes=run_every)
                )
            )

        super(BartenderApp, self).__init__(logger=self.logger, name="BartenderApp")

    def run(self):
        self._startup()

        while not self.stopped():
            for helper_thread in self.helper_threads:
                if not helper_thread.thread.isAlive():
                    self.logger.warning(
                        "%s is dead, restarting" % helper_thread.display_name
                    )
                    helper_thread.start()

            time.sleep(0.1)

        self._shutdown()

    def _startup(self):
        self.logger.info("Starting Bartender...")

        self.logger.info("Verifying message virtual host...")
        self.clients["pyrabbit"].verify_virtual_host()

        self.logger.info("Ensuring admin queue expiration policy...")
        self.clients["pyrabbit"].ensure_admin_expiry()

        self.logger.info("Declaring message exchange...")
        self.clients["pika"].declare_exchange()

        self.logger.info("Starting helper threads...")
        for helper_thread in self.helper_threads:
            helper_thread.start()

        self.logger.info("Loading all local plugins...")
        self.plugin_loader.load_plugins()

        self.logger.info("Starting all local plugins...")
        self.plugin_manager.start_all_plugins()

        try:
            bartender.bv_client.publish_event(name=Events.BARTENDER_STARTED.name)
        except RequestException:
            self.logger.warning("Unable to publish startup notification")

        self.logger.info("Bartender started")

    def _shutdown(self):
        self.logger.info("Shutting down Bartender...")

        self.plugin_manager.stop_all_plugins()

        self.logger.info("Stopping helper threads...")
        for helper_thread in reversed(self.helper_threads):
            helper_thread.stop()

        try:
            bartender.bv_client.publish_event(name=Events.BARTENDER_STOPPED.name)
        except RequestException:
            self.logger.warning("Unable to publish shutdown notification")

        self.logger.info("Successfully shut down Bartender")

    @staticmethod
    def _setup_pruning_tasks():

        prune_tasks = []
        if bartender.config.db.ttl.info > 0:
            prune_tasks.append(
                {
                    "collection": Request,
                    "field": "created_at",
                    "delete_after": timedelta(minutes=bartender.config.db.ttl.info),
                    "additional_query": (
                        Q(status="SUCCESS") | Q(status="CANCELED") | Q(status="ERROR")
                    )
                    & Q(has_parent=False)
                    & Q(command_type="INFO"),
                }
            )

        if bartender.config.db.ttl.action > 0:
            prune_tasks.append(
                {
                    "collection": Request,
                    "field": "created_at",
                    "delete_after": timedelta(minutes=bartender.config.db.ttl.action),
                    "additional_query": (
                        Q(status="SUCCESS") | Q(status="CANCELED") | Q(status="ERROR")
                    )
                    & Q(has_parent=False)
                    & Q(command_type="ACTION"),
                }
            )

        if bartender.config.db.ttl.event > 0:
            prune_tasks.append(
                {
                    "collection": Event,
                    "field": "timestamp",
                    "delete_after": timedelta(minutes=bartender.config.db.ttl.event),
                }
            )

        # Look at the various TTLs to determine how often to run the MongoPruner
        real_ttls = [x for x in bartender.config.db.ttl.values() if x > 0]
        run_every = min(real_ttls) / 2 if real_ttls else None

        return prune_tasks, run_every


class HelperThread(object):
    def __init__(self, init_callable, *args, **kwargs):
        self.logger = logging.getLogger(__name__)

        self.loader_func = partial(init_callable, *args, **kwargs)
        self.thread = None

    def start(self):
        self.thread = self.loader_func()
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        if not self.thread.isAlive():
            self.logger.warning(
                "Uh-oh. Looks like a bad shutdown - the %s " "was already stopped",
                self.display_name,
            )
        else:
            self.logger.debug("%s is being requested to stop", self.display_name)
            self.thread.stop()

            self.logger.debug("Waiting for %s to stop...", self.display_name)
            self.thread.join(2)

            if self.thread.isAlive():
                self.logger.warning("%s did not stop successfully.", self.display_name)
            else:
                self.logger.debug("%s successfully stopped", self.display_name)

    @property
    def display_name(self):
        return getattr(self.thread, "display_name", str(self.thread))



import logging
import random
import string
from datetime import datetime

import mongoengine
import pika.spec
from pyrabbit2.http import HTTPError
from time import sleep

import bartender
import bartender._version
import bg_utils
from bg_utils.mongo.models import Instance, Request, System, StatusInfo
from bg_utils.pika import get_routing_key, get_routing_keys
from brewtils.errors import ModelValidationError, RestError
from brewtils.schema_parser import SchemaParser


class BartenderHandler(object):
    """Implements the BREWMASTER Thrift interface."""

    def __init__(self, registry, clients, plugin_manager, request_validator):
        self.logger = logging.getLogger(__name__)
        self.registry = registry
        self.clients = clients
        self.plugin_manager = plugin_manager
        self.request_validator = request_validator
        self.parser = SchemaParser()

    def processRequest(self, request_id):
        """Validates and publishes a Request.

        :param str request_id: The ID of the Request to process
        :raises InvalidRequest: If the Request is invalid in some way
        :return: None
        """
        request_id = str(request_id)
        self.logger.info("Processing Request: %s", request_id)

        try:
            request = Request.find_or_none(request_id)
            if request is None:
                raise ModelValidationError(
                    "Could not find request with ID '%s'" % request_id
                )

            # Validates the request based on what is in the database.
            # This includes the validation of the request parameters,
            # systems are there, commands are there etc.
            request = self.request_validator.validate_request(request)
            request.save()

            try:
                self.clients["pika"].publish_request(
                    request,
                    confirm=True,
                    mandatory=True,
                    delivery_mode=pika.spec.PERSISTENT_DELIVERY_MODE,
                )
            except Exception:
                msg = "Error while publishing request to queue (%s[%s]-%s %s)" % (
                    request.system,
                    request.system_version,
                    request.instance_name,
                    request.command,
                )
                raise bg_utils.bg_thrift.PublishException(msg)

        except (mongoengine.ValidationError, ModelValidationError, RestError) as ex:
            self.logger.exception(ex)
            raise bg_utils.bg_thrift.InvalidRequest(request_id, str(ex))

    def initializeInstance(self, instance_id):
        """Initializes an instance.

        :param instance_id: The ID of the instance
        :return: QueueInformation object describing message queue for this system
        """
        instance = self._get_instance(instance_id)
        system = self._get_system(instance)

        self.logger.info(
            "Initializing instance %s[%s]-%s",
            system.name,
            instance.name,
            system.version,
        )

        routing_words = [system.name, system.version, instance.name]
        req_name = get_routing_key(*routing_words)
        req_args = {"durable": True, "arguments": {"x-max-priority": 1}}
        req_queue = self.clients["pika"].setup_queue(req_name, req_args, [req_name])

        routing_words.append(
            "".join(
                random.choice(string.ascii_lowercase + string.digits) for _ in range(10)
            )
        )
        admin_keys = get_routing_keys(*routing_words, is_admin=True)
        admin_args = {"durable": True}
        admin_queue = self.clients["pika"].setup_queue(
            admin_keys[-1], admin_args, admin_keys
        )

        connection = {
            "host": bartender.config.publish_hostname,
            "port": bartender.config.amq.connections.message.port,
            "user": bartender.config.amq.connections.message.user,
            "password": bartender.config.amq.connections.message.password,
            "virtual_host": bartender.config.amq.virtual_host,
            "ssl": {"enabled": bartender.config.amq.connections.message.ssl.enabled},
        }

        instance.status = "INITIALIZING"
        instance.status_info = StatusInfo(heartbeat=datetime.utcnow())
        instance.queue_type = "rabbitmq"
        instance.queue_info = {
            "admin": admin_queue,
            "request": req_queue,
            "connection": connection,
            "url": self.clients["public"].connection_url,
        }
        instance.save()

        # Send a request to start to the plugin on the plugin's admin queue
        self.clients["pika"].start(
            system=system.name, version=system.version, instance=instance.name
        )

        return self.parser.serialize_instance(instance, to_string=True)

    def startInstance(self, instance_id):
        """Starts an instance.

        :param instance_id: The ID of the instance
        :return: None
        """
        instance = self._get_instance(instance_id)
        self.plugin_manager.start_plugin(self._get_plugin_from_instance_id(instance_id))
        return self.parser.serialize_instance(instance, to_string=True)

    def stopInstance(self, instance_id):
        """Stops an instance.

        :param instance_id: The ID of the instance
        :return: None
        """
        instance = self._get_instance(instance_id)
        local_plugin = self._get_plugin_from_instance_id(instance_id)
        if local_plugin:
            self.plugin_manager.stop_plugin(local_plugin)
        else:
            system = self._get_system(instance)

            # This causes the request consumer to terminate itself, which ends the plugin
            self.clients["pika"].stop(
                system=system.name, version=system.version, instance=instance.name
            )

        return self.parser.serialize_instance(instance, to_string=True)

    def restartInstance(self, instance_id):
        """Restarts a System. Currently unused."""
        self.stopInstance(instance_id)
        self.startInstance(instance_id)

    def reloadSystem(self, system_id):
        """Reload a system configuration

        :param system_id: The system id
        :return None
        """
        try:
            system = System.objects.get(id=system_id)

            self.logger.info("Reloading system: %s-%s", system.name, system.version)
            self.plugin_manager.reload_system(system.name, system.version)
        except mongoengine.DoesNotExist:
            raise bg_utils.bg_thrift.InvalidSystem(
                "", "Couldn't find system %s" % system_id
            )

    def removeSystem(self, system_id):
        """Removes a system from the registry if necessary.

        :param system_id: The system id
        :return:
        """
        try:
            system = System.objects.get(id=system_id)
        except mongoengine.DoesNotExist:
            raise bg_utils.bg_thrift.InvalidSystem(
                "", "Couldn't find system %s" % system_id
            )

        # Attempt to stop the plugins
        registered = self.registry.get_plugins_by_system(system.name, system.version)

        # Local plugins get stopped by us
        if registered:
            for plugin in registered:
                self.plugin_manager.stop_plugin(plugin)
                self.registry.remove(plugin.unique_name)

        # Remote plugins get a stop request
        else:
            self.clients["pika"].stop(system=system.name, version=system.version)
            count = 0
            while (
                any(instance.status != "STOPPED" for instance in system.instances)
                and count < bartender.config.plugin.local.timeout.shutdown
            ):
                sleep(1)
                count += 1
                system.reload()

        system.reload()

        # Now clean up the message queues
        for instance in system.instances:

            # It is possible for the request or admin queue to be none if we are
            # stopping an instance that was not properly started.
            request_queue = instance.queue_info.get("request", {}).get("name")
            admin_queue = instance.queue_info.get("admin", {}).get("name")

            self.clients["pyrabbit"].destroy_queue(
                request_queue, force_disconnect=(instance.status != "STOPPED")
            )
            self.clients["pyrabbit"].destroy_queue(
                admin_queue, force_disconnect=(instance.status != "STOPPED")
            )

        # Finally, actually delete the system
        system.deep_delete()

    def rescanSystemDirectory(self):
        """Scans plugin directory and starts any new Systems"""
        self.logger.info("Rescanning system directory.")
        self.plugin_manager.scan_plugin_path()

    def getQueueInfo(self, system_name, system_version, instance_name):
        """Gets the size of a queue

        :param system_name: The system name
        :param system_version: The system version
        :param instance_name: The instance name
        :return size of the queue
        :raises Exception: If queue does not exist
        """
        routing_key = get_routing_key(system_name, system_version, instance_name)
        self.logger.debug("Get the queue state for %s", routing_key)

        return bg_utils.bg_thrift.QueueInfo(
            routing_key, self.clients["pyrabbit"].get_queue_size(routing_key)
        )

    def clearQueue(self, queue_name):
        """Clear all Requests in the given queue

        Will iterate through all requests on a queue and mark them as "CANCELED".

        :param queue_name: The queue to clean
        :raises InvalidSystem: If the system_name/instance_name does not match a queue
        """
        try:
            self.logger.debug("Clearing queue %s", queue_name)
            self.clients["pyrabbit"].clear_queue(queue_name)
        except HTTPError as ex:
            if ex.status == 404:
                raise bg_utils.bg_thrift.InvalidSystem(
                    queue_name, "No queue named %s" % queue_name
                )
            else:
                raise

    def clearAllQueues(self):
        """Clears all queues that Bartender knows about.

        :return: None
        """
        self.logger.debug("Clearing all queues")
        systems = System.objects.all()

        for system in systems:
            for instance in system.instances:
                routing_key = get_routing_key(
                    system.name, system.version, instance.name
                )
                self.clearQueue(routing_key)

    def getVersion(self):
        """Gets the current version of the backend"""
        self.logger.debug("Getting Version")

        return bartender._version.__version__

    def ping(self):
        """A simple Method to test connectivity."""
        self.logger.info("Ping.")

    def _get_plugin_from_instance_id(self, instance_id):
        instance = self._get_instance(instance_id)
        system = self._get_system(instance)
        unique_name = self.registry.get_unique_name(
            system.name, system.version, instance.name
        )

        return self.registry.get_plugin(unique_name)

    @staticmethod
    def _get_instance(instance_id):
        try:
            return Instance.objects.get(id=instance_id)
        except mongoengine.DoesNotExist:
            raise bg_utils.bg_thrift.InvalidSystem(
                "", "Couldn't find instance %s" % instance_id
            )

    @staticmethod
    def _get_system(instance):
        try:
            return System.objects.get(instances__contains=instance)
        except mongoengine.DoesNotExist:
            raise bg_utils.bg_thrift.InvalidSystem(
                "", "Couldn't find system " "with instance %s" % instance.id
            )



import re

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()


def find_version(version_file):
    version_line = open(version_file, "rt").read()
    match_object = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_line, re.M)

    if not match_object:
        raise RuntimeError("Unable to find version string in %s" % version_file)

    return match_object.group(1)


setup(
    name="bartender",
    version=find_version("bartender/_version.py"),
    description="Beergarden Backend",
    long_description=readme,
    author="The beer-garden Team",
    author_email="bartender@beer-garden.io",
    url="https://beer-garden.io",
    packages=(find_packages(exclude=["test", "test.*"])),
    license="MIT",
    keywords="bartender beer beer-garden beergarden",
    install_requires=["bg-utils>=2.4.4", "brewtils>=2.4.0", "pyrabbit2==1.0.7"],
    extras_require={
        ':python_version=="2.7"': [
            "future>=0.16.0",
            "futures>=3.1.1",
            "subprocess32>=3.2.7",
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    entry_points={
        "console_scripts": [
            "generate_bartender_config=bartender.__main__:generate_config",
            "migrate_bartender_config=bartender.__main__:migrate_config",
            "generate_bartender_log_config=bartender.__main__:generate_logging_config",
            "bartender=bartender.__main__:main",
        ]
    },
)



from bg_utils.pika import get_routing_key, TransientPikaClient
from brewtils.models import Request
from brewtils.schema_parser import SchemaParser


class PikaClient(TransientPikaClient):
    """Pika client that exposes additional Bartender-specific operations"""

    def publish_request(self, request, **kwargs):
        if "headers" not in kwargs:
            kwargs["headers"] = {}
        kwargs["headers"]["request_id"] = str(request.id)

        if "routing_key" not in kwargs:
            kwargs["routing_key"] = get_routing_key(
                request.system, request.system_version, request.instance_name
            )

        return self.publish(SchemaParser.serialize_request(request), **kwargs)

    def start(self, system=None, version=None, instance=None, clone_id=None):
        self.publish_request(
            Request(
                system=system,
                system_version=version,
                instance_name=instance,
                command="_start",
                command_type="EPHEMERAL",
                parameters={},
            ),
            routing_key=get_routing_key(
                system, version, instance, clone_id, is_admin=True
            ),
        )

    def stop(self, system=None, version=None, instance=None, clone_id=None):
        self.publish_request(
            Request(
                system=system,
                system_version=version,
                instance_name=instance,
                command="_stop",
                command_type="EPHEMERAL",
                parameters={},
            ),
            routing_key=get_routing_key(
                system, version, instance, clone_id, is_admin=True
            ),
        )



import logging
import os
import sys
from threading import Thread
from time import sleep
import signal

import bartender
from bartender.local_plugins.env_help import expand_string_with_environment_var
from mongoengine import DoesNotExist, OperationError

from bartender.local_plugins.logger import getLogLevels, getPluginLogger
from brewtils.stoppable_thread import StoppableThread

# This is the recommended import pattern, see https://github.com/google/python-subprocess32
if os.name == "posix" and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess


class LocalPluginRunner(StoppableThread):
    """Class for running a local plugin in its own process.

    Can be stopped/started and killed like a normal process"""

    def __init__(
        self,
        entry_point,
        system,
        instance_name,
        path_to_plugin,
        web_host,
        web_port,
        ssl_enabled,
        plugin_args=None,
        environment=None,
        requirements=None,
        plugin_log_directory=None,
        url_prefix=None,
        ca_verify=True,
        ca_cert=None,
        **kwargs
    ):
        self.entry_point = entry_point
        self.system = system
        self.instance_name = instance_name
        self.path_to_plugin = path_to_plugin
        self.web_host = web_host
        self.web_port = web_port
        self.ssl_enabled = ssl_enabled
        self.plugin_args = plugin_args or []
        self.environment = environment or {}
        self.requirements = requirements or []
        self.plugin_log_directory = plugin_log_directory
        self.url_prefix = url_prefix
        self.ca_verify = ca_verify
        self.ca_cert = ca_cert
        self.username = kwargs.get("username", None)
        self.password = kwargs.get("password", None)
        self.plugin_default_log_level = kwargs.get("log_level", logging.INFO)

        for instance in self.system.instances:
            if instance.name == self.instance_name:
                self.instance = instance
                break

        self.process = None
        self.executable = [sys.executable]
        if self.entry_point.startswith("-m "):
            self.executable.append("-m")
            self.executable.append(self.entry_point.split(" ", 1)[1])
        else:
            self.executable.append(self.entry_point)
        self.executable += self.plugin_args

        self.unique_name = "%s[%s]-%s" % (
            self.system.name,
            self.instance_name,
            self.system.version,
        )

        self.log_levels = getLogLevels()
        log_config = {
            "log_directory": self.plugin_log_directory,
            "log_name": self.unique_name,
        }

        # Logger used for bartender purposes.
        self.logger = getPluginLogger(
            self.unique_name,
            format_string="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            **log_config
        )

        log_config["log_level"] = self.plugin_default_log_level
        self.unformatted_logger = getPluginLogger(
            self.unique_name + "-uf", **log_config
        )

        StoppableThread.__init__(self, logger=self.logger, name=self.unique_name)

    @property
    def status(self):
        try:
            # TODO: Remove this reload. We have to find a better way to store the status
            # that doesn't require a database connection.
            self.instance.reload()
            return self.instance.status
        except (DoesNotExist, OperationError):
            self.logger.error("Error getting status of plugin %s" % self.unique_name)
            return "UNKNOWN"

    @status.setter
    def status(self, value):
        try:
            # TODO: Remove this reload. We have to find a better way to store the status
            # that doesn't require a database connection.
            self.instance.reload()
            self.instance.status = value
            self.instance.save()
        except (DoesNotExist, OperationError):
            self.logger.error(
                "Error updating status of plugin %s to %s" % (self.unique_name, value)
            )

    def kill(self):
        """Kills the plugin by killing the underlying process."""
        if self.process and self.process.poll() is None:
            self.logger.warning("About to kill plugin %s", self.unique_name)
            self.process.kill()
            self.logger.warning("Plugin %s has been killed", self.unique_name)

    def run(self):
        """Runs the plugin

        Run the plugin using the entry point specified with the generated environment in its own
        subprocess. Pipes STDOUT and STDERR such that when the plugin stops executing
        (or IO is flushed) it will log it.
        """
        try:
            self.logger.info(
                "Starting plugin %s subprocess: %s", self.unique_name, self.executable
            )
            self.process = subprocess.Popen(
                self.executable,
                bufsize=0,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                env=self._generate_plugin_environment(),
                cwd=os.path.abspath(self.path_to_plugin),
                preexec_fn=lambda: signal.signal(signal.SIGINT, signal.SIG_IGN),
            )

            # Reading the process IO is blocking and we need to shutdown
            # gracefully, so reading IO needs to be its own thread
            stdout_thread = Thread(
                target=self._check_io,
                name=self.unique_name + "_stdout_thread",
                args=(self.process.stdout, logging.INFO),
            )
            stderr_thread = Thread(
                target=self._check_io,
                name=self.unique_name + "_stderr_thread",
                args=(self.process.stderr, logging.ERROR),
            )
            stdout_thread.start()
            stderr_thread.start()

            # Just spin here until until the process is no longer alive
            while self.process.poll() is None:
                sleep(0.1)

            self.logger.info(
                "Plugin %s subprocess has stopped with exit status %s, "
                "performing final IO read(s)",
                self.unique_name,
                self.process.poll(),
            )
            stdout_thread.join()
            stderr_thread.join()

            # If stopped wasn't set then this was not expected
            if not self.stopped():
                self.logger.error("Plugin %s unexpectedly shutdown!", self.unique_name)

            self.logger.info("Plugin %s is officially stopped", self.unique_name)

        except Exception as ex:
            self.logger.error("Plugin %s died", self.unique_name)
            self.logger.error(str(ex))

    def _check_io(self, stream, default_level):
        """Helper function thread target to read IO from the plugin's subprocess

        This will read line by line from STDOUT or STDERR. If the line includes
        one of the log levels that the python logger knows about we assume that
        the plugin has its own logger and formatter. In that case we log to our
        unformatted logger at the same level to keep the original formatting.

        If we can't determine the log level then we'll log the message at
        ``default_level``. That way we guarantee messages are outputted (this
        is usually caused by a plugin writing to STDOUT / STDERR directly or
        raising an exception with a stacktrace).
        """
        stream_reader = iter(stream.readline, "")

        for raw_line in stream_reader:
            line = raw_line.rstrip()

            level_to_log = default_level
            for level in self.log_levels:
                if line.find(level) != -1:
                    level_to_log = getattr(logging, level)
                    break

            self.unformatted_logger.log(level_to_log, line)

        if self.process.poll() is None:
            self.logger.debug("Process isn't quite dead yet, reading IO again")
            self._check_io(stream, default_level)

    def _generate_plugin_environment(self):

        plugin_env = {
            "BG_NAME": self.system.name,
            "BG_VERSION": self.system.version,
            "BG_INSTANCE_NAME": self.instance_name,
            "BG_PLUGIN_PATH": self.path_to_plugin,
            "BG_WEB_HOST": self.web_host,
            "BG_WEB_PORT": self.web_port,
            "BG_SSL_ENABLED": self.ssl_enabled,
            "BG_URL_PREFIX": self.url_prefix,
            "BG_CA_VERIFY": self.ca_verify,
            "BG_CA_CERT": self.ca_cert,
            "BG_USERNAME": self.username,
            "BG_PASSWORD": self.password,
            "BG_LOG_LEVEL": logging.getLevelName(self.plugin_default_log_level),
        }

        for key, value in plugin_env.items():
            plugin_env[key] = str(value)

        # Allowed host env vars
        for env_var in bartender.config.plugin.local.host_env_vars:
            if env_var in plugin_env:
                self.logger.warning(
                    "%s is present in the list of host environment variables to "
                    "propagate to local plugins, but that environment variable name is "
                    "reserved by Beergarden for internal use. The host value will NOT "
                    "be used." % env_var
                )
            else:
                plugin_env[env_var] = os.environ.get(env_var, "")

        for key, value in self.environment.items():
            plugin_env[key] = expand_string_with_environment_var(str(value), plugin_env)

        return plugin_env
